# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""Main file for scheduler options related operations on the commcell.

This file has all the classes related to Schedule Options operations.

ScheduleOptions: Class for getting Schedule Options for Schedule and Schedule Policies

BackupOptions: Class for Backup Options for Schedule and Schedule Policies

AuxCopyOptions: Class for AuxCopy Options for Schedule and Schedule Policies

ScheduleOptions:

    __new__()       -- Returns the respective class object based on the option_type

    __init__()      --  initialises ScheduleOptions for Schedule and Schedule Policies

    options_json()  --  Returns the options json for the new options provided

BackupOptions:

    __init__()      --  initialises BackupOptions for Schedule and Schedule Policies

    options_json()  --  Returns the options json for the new backup options provided

AuxCopyOptions:

    __init__()      --  initialises AuxCopyOptions for Schedule and Schedule Policies

    options_json()  --  Returns the options json for the new AuxCopy options provided

"""

from abc import ABCMeta, abstractmethod

class ScheduleOptions:
    """Class for getting Schedule Options for Schedule and Schedule Policies."""

    __metaclass__ = ABCMeta

    #This map has to be updated with the subdict option_name for every new option class added
    policy_to_options_map = {

        'Data Protection': 'backupOpts',
        'Auxiliary Copy': 'auxcopyJobOption'
    }

    def __new__(cls, options_type, current_options=None):
        """
        Returns the respective class object based on the option_type

        Args:
            options_type (str) -- option type as per ScheduleOptions.options dict.
            current_options (dict) -- current options set for the schedule if any.

        Returns (obj) -- Return the class object based on the option_type

        """

        #This dict has to be update with the option_name and corresponding Option class created
        options = {

            'backupOpts': BackupOptions,
            'auxcopyJobOption': AuxCopyOptions

        }
        #subclass inherit __new__ method so we need this if check to initialize parent.
        if cls is not __class__:
            return super().__new__(cls)
        return options[options_type](current_options)




    def __init__(self, options_type, current_options=None):
        """
        Initialises Schedule Options class
        Args:
            options_type (str) -- option type as per ScheduleOptions.options dict.
            current_options (dict) -- current options set for the schedule if any.
        """
        if current_options:
            self.current_options = current_options
        else:
            self.current_options = {}

    @abstractmethod
    def options_json(self, new_options=None):
        """
        Returns the options json for the new options provided
        Args:
            new_options: options_json based on the type of scheduler option

        Returns (dict) -- new options

        """


class BackupOptions(ScheduleOptions):
    """Class for getting Backup Schedule Options for Schedule and Schedule Policies."""
    def __init__(self, options_type, current_options=None):
        """
        Initialises the BackupOptions class
        Args:
            options_type (str) -- should be 'backupOpts'
            current_options (dict) -- current backup options set for the schedule if any.
        """
        super().__init__(options_type, current_options)

    def options_json(self, new_options=None):
        """

        Returns the backup options json for the new options provided

        Args:
             new_options (dict) -- options which need to be set for the schedule
                                    Example:
                                     {
                                        "backupLevel": backup_level(Full / Incremental / Differential / Synthetic_full),
                                        "incLevel": incremental_level(BEFORE_SYNTH / AFTER_SYNTH),
                                        "runIncrementalBackup": incremental_backup
                                    }
        Returns (dict) -- new options


        """
        if not new_options:
            new_options = {}
        if self.current_options:
            for key, value in new_options.items():
                self.current_options[key] = value
            return {'backupOpts': self.current_options}

        default_dict = {
            "backupLevel": "Incremental",
            "incLevel": 1,
            "runIncrementalBackup": False
        }

        new_options = dict(default_dict, **new_options)
        return {'backupOpts': new_options}


class AuxCopyOptions(ScheduleOptions):
    """Class for getting AuxCopy Schedule Options for Schedule and Schedule Policies."""
    def __init__(self, options_type, current_options=None):
        """
        Initialises the AuxCopyOptions class
        Args:
            options_type (str) -- should be 'auxcopyJobOption'
            current_options (dict) -- current AuxCopy options set for the schedule if any.
        """
        super().__init__(options_type, current_options)

    def options_json(self, new_options=None):
        """

        Returns the AuxCopy options json for the new options provided

        Args:
             new_options (dict) -- options which need to be set for the schedule
                                    Example:
                                    {
                                        "maxNumberOfStreams": 0,
                                        "useMaximumStreams": True,
                                        "useScallableResourceManagement": True,
                                        "totalJobsToProcess": 1000,
                                        "allCopies": True,
                                        "mediaAgent": {
                                            "mediaAgentName": "<ANY MEDIAAGENT>"
                                        }
                                    }
        """
        if not new_options:
            new_options = {}
        if self.current_options:
            for key, value in new_options.items():
                self.current_options[key] = value
            return {'backupOpts':
                        {
                            'mediaOpt':
                                {
                                    'auxcopyJobOption': self.current_options
                                }
                        }
                    }

        default_dict = {
            "maxNumberOfStreams": 0,
            "useMaximumStreams": True,
            "useScallableResourceManagement": True,
            "totalJobsToProcess": 1000,
            "allCopies": True,
            "mediaAgent": {
                "mediaAgentName": "<ANY MEDIAAGENT>"
            }
        }

        new_options = dict(default_dict, **new_options)

        return {
            'backupOpts':
            {
                'mediaOpt':
                    {
                        'auxcopyJobOption': new_options
                    }
            }
            }



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""File for operating on a MYSQL Subclient

MYSQLSubclient is the only class defined in this file.

MYSQLSubclient: Derived class from Subclient Base class, representing a MYSQL subclient,
                        and to perform operations on that subclient

MYSQLSubclient:
    __init__()                          --  constructor for the class

    is_failover_to_production()         --  Sets the isFailOverToProduction flag for the
    subclient as the value provided as input

    _backup_request_json()              --  prepares the json for the backup request

    _get_subclient_properties()         --  Gets the subclient related properties of MYSQL subclient

    _get_subclient_properties_json()    --  get the all subclient related properties of this
    subclient

    content()                           --  Creates the list of content JSON to pass to the API to
    add/update content of a MYSQL Subclient

    backup()                            --  Runs a backup job for the subclient of the level
    specified

    restore_in_place()                  --  Restores the mysql data/log files specified in
    the input paths list to the same location


MYSQLSubclient instance Attributes:
===================================

    **is_blocklevel_backup_enabled**    --  Returns True if block level backup is
    enabled else returns false

    **is_proxy_enabled**                --  Returns True if proxy is enabled in the subclient

    **is_failover_to_production**       --  Returns the isFailOverToProduction flag of the subclient

    **content**                         --  Returns the appropriate content from
    the Subclient relevant to the user

"""

from __future__ import unicode_literals
from ..subclient import Subclient
from ..exception import SDKException


class MYSQLSubclient(Subclient):
    """Derived class from Subclient Base class, representing a MYSQL subclient,
        and to perform operations on that subclient."""

    def __init__(self, backupset_object, subclient_name, subclient_id=None):
        """Initialise the Subclient object.

            Args:
                backupset_object (object)  --  instance of the Backupset class

                subclient_name   (str)     --  name of the subclient

                subclient_id     (str)     --  id of the subclient
                    default: None

            Returns:
                object - instance of the MYSQLSubclient class

        """
        self.mysql_subclient_prop = None
        self.dfs_subclient_prop = None
        self.plan_entity = None
        self.cassandra_props = None
        self.analytics_subclient_prop = None
        super(MYSQLSubclient, self).__init__(backupset_object, subclient_name, subclient_id)

    @property
    def is_blocklevel_backup_enabled(self):
        """returns True if block level backup is enabled else returns false

        Returns:
            (bool) - boolean value based on blocklevel enable status

                    True if block level is enabled
                    False if block level is not enabled

        """
        return bool(self._subclient_properties.get(
            'mySqlSubclientProp', {}).get('isUseBlockLevelBackup', False))

    @property
    def is_proxy_enabled(self):
        """Returns True if proxy is enabled in the subclient

        Returns:
            (bool) - boolean value based on proxy enable status

                    True if proxy is enabled
                    False if proxy is not enabled

        """
        return self._subclient_properties.get(
            'mySqlSubclientProp', {}).get('proxySettings', {}).get(
                'isProxyEnabled', False)

    @property
    def is_failover_to_production(self):
        """Returns the isFailOverToProduction flag of the subclient.

        Returns:

            (bool)  --  True if flag is set
                        False if the flag is not set

        """
        return self._subclient_properties.get(
            'mySqlSubclientProp', {}).get(
                'proxySettings', {}).get('isFailOverToProduction', False)

    @is_failover_to_production.setter
    def is_failover_to_production(self, value):
        """Sets the isFailOverToProduction flag for the subclient as the value provided as input.

        Args:

            value   (bool)  --  Boolean value to set as flag

            Raises:
                SDKException:
                    if failed to set isFailOverToProduction flag

                    if the type of value input is not bool
        """
        if isinstance(value, bool):
            self._set_subclient_properties(
                "_subclient_properties['mySqlSubclientProp']\
                ['proxySettings']['isFailOverToProduction']",
                value)
        else:
            raise SDKException(
                'Subclient', '102', 'Expecting a boolean value here'
            )

    def _backup_request_json(
            self,
            backup_level,
            inc_with_data=False,
            truncate_logs_on_source=False,
            do_not_truncate_logs=False,
            schedule_pattern=None
    ):
        """
        prepares the json for the backup request

            Args:
                backup_level            (list)  --  level of backup the user wish to run

                    Accepted Values:
                        Full / Incremental / Differential

                inc_with_data           (bool)  --  flag to determine if the incremental backup
                includes data or not

                truncate_logs_on_source (bool)  --  flag to determine if the logs to be
                truncated on master client

                    default: False

                do_not_truncate_logs    (bool)  --  flag to determine if the proxy logs
                needs to be truncated or not

                    default: False

                schedule_pattern (dict) -- scheduling options to be included for the task

                        Please refer schedules.schedulePattern.createSchedule()
                                                                    doc for the types of Jsons

            Returns:
                dict - JSON request to pass to the API

        """
        request_json = self._backup_json(backup_level, False, "BEFORE_SYNTH", schedule_pattern=schedule_pattern)

        backup_options = {
            "truncateLogsOnSource":truncate_logs_on_source,
            "sybaseSkipFullafterLogBkp":False,
            "notSynthesizeFullFromPrevBackup":False,
            "incrementalDataWithLogs":inc_with_data,
            "backupLevel":backup_level,
            "incLevel":"NONE",
            "adHocBackup":False,
            "runIncrementalBackup":False,
            "doNotTruncateLog":do_not_truncate_logs,
            "dataOpt":{
                "skipCatalogPhaseForSnapBackup":True,
                "createBackupCopyImmediately":True,
                "useCatalogServer":True,
                "followMountPoints":False,
                "enforceTransactionLogUsage":False,
                "skipConsistencyCheck":False,
                "createNewIndex":False
            },
            "mediaOpt":{

            }
        }
        request_json["taskInfo"]["subTasks"][0]["options"][
            "backupOpts"] = backup_options

        return request_json

    def _get_subclient_properties(self):
        """Gets the subclient related properties of MYSQL subclient"""
        super(MYSQLSubclient, self)._get_subclient_properties()
        if 'mySqlSubclientProp' in self._subclient_properties:
            self.mysql_subclient_prop = self._subclient_properties['mySqlSubclientProp']
        if 'dfsSubclientProp' in self._subclient_properties:
            self.dfs_subclient_prop = self._subclient_properties['dfsSubclientProp']
        if 'planEntity' in self._subclient_properties:
            self.plan_entity = self._subclient_properties['planEntity']
        if 'cassandraProps' in self._subclient_properties:
            self.cassandra_props = self._subclient_properties['cassandraProps']
        if 'content' in self._subclient_properties:
            self._content = self._subclient_properties['content']
        if 'analyticsSubclientProp' in self._subclient_properties:
            self.analytics_subclient_prop = self._subclient_properties['analyticsSubclientProp']

    def _get_subclient_properties_json(self):
        """get the all subclient related properties of this subclient.

           Returns:
                dict - all subclient properties put inside a dict

        """
        subclient_json = {
            "subClientProperties":
                {
                    "proxyClient": self._proxyClient,
                    "mySqlSubclientProp": self.mysql_subclient_prop,
                    "subClientEntity": self._subClientEntity,
                    "dfsSubclientProp": self.dfs_subclient_prop,
                    "planEntity": self.plan_entity,
                    "cassandraProps": self.cassandra_props,
                    "content": self._content,
                    "commonProperties": self._commonProperties,
                    "analyticsSubclientProp": self.analytics_subclient_prop,
                    "contentOperationType": 1
                }
        }
        return subclient_json

    @property
    def content(self):
        """Returns the appropriate content from the Subclient relevant to the user.

            Returns:
                list - list of content associated with the subclient

        """
        cont = []

        # Getting the database names from subclient content details
        for path in self._content:
            for key, value in path.items():
                if key == "mySQLContent":
                    cont.append(value["databaseName"])
        return cont

    @content.setter
    def content(self, subclient_content):
        """Creates the list of content JSON to pass to the API to add/update content of a
            MYSQL Subclient.

            Args:
                subclient_content (list)  --  list of the content to add to the subclient

            Returns:
                list - list of the appropriate JSON for an agent to send to the POST Subclient API

        """
        cont = []
        for mysql_cont in subclient_content:
            mysql_dict = {
                "mySQLContent": {
                    "databaseName": mysql_cont
                }
            }
            cont.append(mysql_dict)

        self._set_subclient_properties("_content", cont)


    def backup(
            self,
            backup_level="Differential",
            inc_with_data=False,
            truncate_logs_on_source=False,
            do_not_truncate_logs=False,
            schedule_pattern=None
    ):
        """Runs a backup job for the subclient of the level specified.

            Args:
                backup_level        (str)   --  level of backup the user wish to run
                        Full / Incremental / Differential / Synthetic_full

                    default: Differential

                inc_with_data       (bool)  --  flag to determine if the incremental backup
                includes data or not

                truncate_logs_on_source (bool)  --  flag to determine if the logs to be
                truncated on master client

                    default: False

                do_not_truncate_logs    (bool)  --  flag to determine if the proxy logs
                needs to be truncated or not

                    default: False

                schedule_pattern (dict) -- scheduling options to be included for the task

                        Please refer schedules.schedulePattern.createSchedule()
                                                                    doc for the types of Jsons

            Returns:
                object - instance of the Job class for this backup job if its an immediate Job

                         instance of the Schedule class for the backup job if its a scheduled Job

            Raises:
                SDKException:
                    if backup level specified is not correct

                    if response is empty

                    if response is not success

        """
        backup_level = backup_level.lower()

        if backup_level not in ['full', 'incremental', 'differential', 'synthetic_full']:
            raise SDKException('Subclient', '103')

        if not (inc_with_data or truncate_logs_on_source or do_not_truncate_logs or schedule_pattern):
            return super(MYSQLSubclient, self).backup(backup_level)
        request_json = self._backup_request_json(
            backup_level,
            inc_with_data,
            truncate_logs_on_source=truncate_logs_on_source,
            do_not_truncate_logs=do_not_truncate_logs,
            schedule_pattern=schedule_pattern
        )
        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'POST', self._commcell_object._services['CREATE_TASK'], request_json
        )
        return self._process_backup_response(flag, response)

    def restore_in_place(
            self,
            paths=None,
            staging=None,
            dest_client_name=None,
            dest_instance_name=None,
            data_restore=True,
            log_restore=False,
            overwrite=True,
            copy_precedence=None,
            from_time=None,
            to_time=None,
            media_agent=None,
            table_level_restore=False,
            clone_env=False,
            clone_options=None,
            redirect_enabled=False,
            redirect_path=None,
            browse_jobid=None):
        """Restores the mysql data/log files specified in the input paths list to the same location.

            Args:
                paths               (list)  --  list of database/databases to be restored

                staging             (str)   --  staging location for mysql logs during restores

                dest_client_name    (str)   --  destination client name where files are
                                                        to be restored

                dest_instance_name  (str)   --  destination mysql instance name of
                                                        destination client

                data_restore        (bool)  --  for data only/data+log restore

                log_restore         (bool)  --  for log only/data+log restore

                overwrite           (bool)  --  unconditional overwrite files during restore
                    default: True

                copy_precedence     (int)   --  copy precedence value of storage policy copy
                    default: None

                from_time           (str)   --  time to retore the contents after
                        format: YYYY-MM-DD HH:MM:SS

                    default: None

                to_time             (str)   --  time to retore the contents before
                        format: YYYY-MM-DD HH:MM:SS

                    default: None

                media_agent             (str)   --  media agent associated

                    default: None

                table_level_restore     (bool)  --  Table level restore flag

                    default: False

                clone_env               (bool)  --  boolean to specify whether the database
                should be cloned or not

                    default: False

                clone_options           (dict)  --  clone restore options passed in a dict

                    default: None

                    Accepted format: {
                                        "stagingLocaion": "/gk_snap",
                                        "forceCleanup": True,
                                        "port": "5595",
                                        "libDirectory": "",
                                        "isInstanceSelected": True,
                                        "reservationPeriodS": 3600,
                                        "user": "",
                                        "binaryDirectory": "/usr/bin"

                                     }

                redirect_enabled         (bool)  --  boolean to specify if redirect restore is
                enabled

                    default: False

                redirect_path           (str)   --  Path specified in advanced restore options
                in order to perform redirect restore

                    default: None

                browse_jobid           (int)   --  Browse jobid to browse and restore from

                    default: None

            Returns:
                object - instance of the Job class for this restore job

            Raises:
                SDKException:
                    if paths is not a list

                    if failed to initialize job

                    if response is empty

                    if response is not success

        """
        if not (isinstance(paths, list) and
                isinstance(overwrite, bool)):
            raise SDKException('Subclient', '101')

        if paths == []:
            raise SDKException('Subclient', '104')

        instance_object = self._backupset_object._instance_object
        if dest_client_name is None:
            dest_client_name = instance_object._agent_object._client_object.client_name

        if dest_instance_name is None:
            dest_instance_name = instance_object.instance_name
        instance_object._restore_association = self._subClientEntity

        return instance_object.restore_in_place(
            path=paths,
            staging=staging,
            dest_client_name=dest_client_name,
            dest_instance_name=dest_instance_name,
            data_restore=data_restore,
            log_restore=log_restore,
            overwrite=overwrite,
            copy_precedence=copy_precedence,
            from_time=from_time,
            to_time=to_time,
            media_agent=media_agent,
            table_level_restore=table_level_restore,
            clone_env=clone_env,
            clone_options=clone_options,
            redirect_enabled=redirect_enabled,
            redirect_path=redirect_path,
            browse_jobid=browse_jobid
        )



# -*- coding: utf-8 -*-
# pylint: disable=R1705, R0205

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------
"""File for operating on an AD agent Backupset.

adbackupset is the only class defined in this file.

Function:
    azuread_browse_double_query     create browse options for objects in folder

    azuread_browse_double_query_adv    create browse options for objects attrbute

    azuread_browse_options_builder    build browse options for azure ad browse
Class:

    AzureADBackupset:  Derived class from Backuset Base class, representing a
                            Azure AD agent backupset, and to perform operations on that backupset

        _azuread_browse_basic : Do basic browse with option

        _azuread_browse_meta  : Get Azure ad folder meta information

        _azuread_browse_folder : Get Azure objects based on the folder type

        _adv_attributes : Get Azure AD object attribute

        browse() : Overwrite default browse operation

        _process_browse_repsonse : process the browse result

        azuread_get_metadata : create azure ad object meta data information
"""

from __future__ import unicode_literals

from ..backupset import Backupset
from ..exception import SDKException


class AzureAdBackupset(Backupset):
    """ Azure AD agent backupset class """

    def _azuread_browse_basic(self, options):
        """ do basic browse activity with options
            Args:
                options    (dict)    browse option from impoort
                                adv_attributes    get object advanced attribute
            Return:
                count     (int)    return count from browse
                result    (list)    objects list from browse
            Raise:
                101    If browse return Nothing
        """
        options = self._prepare_browse_options(options)

        request_json = self._prepare_browse_json(options)
        if "adv_attributes" in options:
            request_json = self.azuread_browse_double_query_adv(options, request_json)
        else:
            request_json = self.azuread_browse_double_query(options, request_json)
        flag, response = self._cvpysdk_object.make_request("POST",
                                                           self._BROWSE,
                                                           request_json)
        if flag:
            count, result = self._process_browse_response(flag, response, options)
        else:
            raise SDKException('Response', '101', self._update_response_(response.text))
        return count, result

    def _azuread_browse_meta(self, options):
        """ get basic browse related meta data
            Args:
                options    (dict)    browse option from impoort
            Return:
                azure_meta    (dict)    azure ad meta data from browse result
            Raise:
                None
        """
        if "meta" in options:
            azure_meta = options['meta']
        else:
            azure_meta = {}
        _, result = self._azuread_browse_basic(options)
        if len(result) == 1 and result[0]['objType'] == 1:
            newid = result[0]['commonData']['id']
            options['filters'] = [(_[0], newid, _[2]) for _ in options['filters']]
            name, metainfo = self.azuread_browse_obj_meta(result[0])
            azure_meta['root'] = metainfo
            options['meta'] = azure_meta
            result = self.browse(**options)
        else:
            for _ in result:
                name, metainfo = self.azuread_browse_obj_meta(_)
                azure_meta[name] = metainfo
        return azure_meta

    def _azuread_browse_folder(self, options):
        """ browse folder content
            Args:
                options    (dict)    browse option from impoort
            Return:
                count     (int)    return count from browse
                result    (list)    objects list from browse
            Raise:
                None
        """
        azure_meta_mapper = {
            "user" : "users",
            "group" : "groups",
            "reg_app" : "applications",
            "ent_app" : "servicePrincipals"}

        azure_meta = options['meta']
        newid = azure_meta[azure_meta_mapper[options['folder']]]['id']

        options['filters'] = [(_[0], newid, _[2]) for _ in options['filters']]
        if "search" in options:
            options['filters'] = [[_, ("30", options['search'], None)] for _ in options['filters']]
        del(options['folder'])
        count, results = self._azuread_browse_basic(options)
        result = []
        for _ in results:
            _, metainfo = self.azuread_browse_obj_meta(_)
            result.append(metainfo)
        if 'adv_attributes' in options:
            count, result = self._adv_attributes(options, count, results)
        return count, result

    def _adv_attributes(self, options, count, results):
        """ get advanced attribute in objects
            Args:
                options    (dict)    browse option from impoort
                count     (int)    return count from browse
                result    (list)    objects list from browse
            Return:
                count     (int)    return count from browse
                result    (list)    objects list from browse
            Raise:
                None
        """
        for _ in results:
            _, obj_ = self.azuread_browse_obj_meta(_)
            rev_azuread_index_meta_mapper = {2 : "USER",
                                             5 : "APPLICATION",
                                             3 : "GROUP",
                                             6 : "SERVICE_PRINCIPAL"}
            options['filters'] = [[("130", obj_['id'], "9"),
                                   ("125", rev_azuread_index_meta_mapper[obj_['type']],
                                    None)],
                                  [("130", obj_['id'], "9"),
                                   ("125", rev_azuread_index_meta_mapper[obj_['type']],
                                    None)]]
            count, results = self._azuread_browse_basic(options)
            result = []
            for _ in results:
                _, metainfo = self.azuread_browse_obj_meta(_)
                result.append(metainfo)
        return count, result

    def browse(self, *args, **kwargs):
        """Browses the content of the Backupset.
            Args:
                args    list    args passed for browse
                kwargs    dict    dict passed for browse
            Return:
                count     (int)    return count from browse
                browse_result    (list)    objects list from browse
            Raise:
                None
        """
        if args and isinstance(args[0], dict):
            options = args[0]
        else:
            options = kwargs

        options = self.azuread_browse_options_builder(options)
        azure_meta = self._azuread_browse_meta(options)
        options['meta'] = azure_meta

        if "folder" in options:
            count, browse_result = self._azuread_browse_folder(options)
        else:
            browse_result = azure_meta
            count = 0
        return count, browse_result

    def _process_browse_response(self, flag, response, options):
        """Retrieves the items from browse response.
            Args:
                flag    (bool)  --  boolean, whether the response was success or not
                response (dict)  --  JSON response received for the request from the Server
                options  (dict)  --  The browse options dictionary
            Returns:
                list - List of only the file / folder paths from the browse response
                dict - Dictionary of all the paths with additional metadata retrieved from browse
            Raises:
                SDKException:
                    if failed to browse/search for content
                    if response is empty
                    if response is not success
        """
        metadatas = []
        if flag:
            if response.json():
                response_json = response.json()
                if response_json and 'browseResponses' in response_json:
                    _browse_responses = response_json['browseResponses']
                    for browse_response in _browse_responses:
                        if "browseResult" in browse_response:
                            browse_result = browse_response['browseResult']
                            browseresulttcount = browse_result['totalItemsFound']
                            if 'dataResultSet' in browse_result:
                                result_set = browse_result['dataResultSet']
                    if not result_set:
                        raise SDKException('Backupset', '102',
                                           "Failed to browse for subclient backup content")
                    else:
                        for result in result_set:
                            metadata = self.azuread_get_metadata(result)
                            metadatas.append(metadata['azureADDataV2'])
            else:
                raise SDKException("Backupset", "102", "response is not valid")
        else:
            raise SDKException('Response', '101', self._update_response_(response.text))
        return browseresulttcount, metadatas

    def azuread_get_metadata(self, result):
        """ Get azure ad meta data for browse result
            Args:
                result    (list)    objects list from browse
            Return:
                metadata    (dict)    azure ad browse meta data
            Raise:
                110    can't find meta data
        """
        metadata = {}
        if "advancedData" in result:
            if "browseMetaData" in result['advancedData']:
                metadata = result['advancedData']['browseMetaData']
                if "azureADDataV2" in metadata:
                    metadata['azureADDataV2']['guid'] = result['advancedData']['objectGuid']
                else:
                    raise SDKException('Backupset', '110',
                                       "Azure AD meta data is not found")
        return metadata

    def azuread_browse_obj_meta(self, obj_):
        """ get azure ad obj meta info
            Args:
                obj_    (obj)    azuare ad object
            Return:
                name    (str)    azure ad display name
                metainfo     (dict)    azure ad browse meta data
            Raise:
                None
        """
        name = obj_['commonData']['displayName']
        metainfo = {}
        metainfo['id'] = obj_['commonData']['id']
        metainfo['azureid'] = metainfo['id'].replace("x", "-")
        metainfo['name'] = name
        metainfo['guid'] = obj_['guid']
        metainfo['type'] = obj_['objType']
#        rev_azuread_index_meta_mapper = dict((v,k) for k,v in AZUREAD_INDEX_META_MAPPER.items())
        azuread_index_meta_mapper = {5 : "application",
                                     3 : "group",
                                     6 : "servicePrincipal",
                                     2 : "user"}
        if metainfo['type'] in azuread_index_meta_mapper:
            metainfo["adv_attribute"] = obj_["{}Data".\
                                             format(azuread_index_meta_mapper[metainfo['type']])]
        return name, metainfo

    def azuread_browse_double_query(self, options, request_json):
        """ create request json for azure ad based on double query
            Args:
                options    (dict)    browse option from impoort
                request_json    (json)    request json file from basic request class
            Return:
                request_json    (json)    request json with addittional options
            Raise:
                None
        """
        if options['filters']:
            request_json['queries'] = []
            for ind_, filter_ in enumerate(options['filters']):
                if ind_ > 0:
                    queries = {
                        "type": ind_,
                        "queryId": str(ind_),
                        }
                    request_json['queries'].append(queries)
                    request_json['queries'][ind_]['agprParam'] = {'aggrType': 4,
                                                                  'field' : 0}
                else:
                    queries = {
                        "type": ind_,
                        "queryId": str(ind_),
                        "dataParam": {
                            "sortParam": {
                                "ascending": True,
                                "sortBy": [126]
                                },
                            "paging": {
                                "pageSize": int(options['page_size']),
                                "skipNode": int(options['skip_node']),
                                "firstNode": 0
                                }
                            }
                        }
                    request_json['queries'].append(queries)
                request_json['queries'][ind_]['whereClause'] = []
                if len(filter_) == 2:
                    for _ in filter_:
                        filter_dict = {
                            'connector': 0,
                            'criteria': {
                                'field': _[0],
                                'values': [_[1]]}}
                        request_json['queries'][ind_]['whereClause'].append(filter_dict)
                else:
                    filter_dict = {
                        'connector': 0,
                        'criteria': {
                            'field': filter_[0],
                            'values': [filter_[1]],
                            "dataOperator" : int(filter_[2])}}
                    request_json['queries'][ind_]['whereClause'].append(filter_dict)
        return request_json


    def azuread_browse_double_query_adv(self, options, request_json):
        """ get advanced attribute for obeject
            Args:
                options    (dict)    browse option from impoort
                request_json    (json)    request json file from basic request class
            Return:
                request_json    (json)    request json with addittional options
            Raise:
                None
        """
        if options['filters']:
            request_json['queries'] = []
            for ind_, filter_ in enumerate(options['filters']):
                if ind_ > 0:
                    queries = {
                        "type": ind_,
                        "queryId": str(ind_),
                        }
                    request_json['queries'].append(queries)
                    request_json['queries'][ind_]['agprParam'] = {'aggrType': 4,
                                                                  'field' : 0}
                else:
                    queries = {
                        "type": ind_,
                        "queryId": str(ind_),
                        "dataParam": {
                            "sortParam": {
                                "ascending": True,
                                "sortBy": [126]
                                },
                            "paging": {
                                "pageSize": int(options['page_size']),
                                "skipNode": int(options['skip_node']),
                                "firstNode": 0
                                }
                            }
                        }
                    request_json['queries'].append(queries)
                request_json['queries'][ind_]['whereClause'] = []

                if len(filter_) == 2:
                    for _ in filter_:
                        if _[2] is not None:
                            filter_dict = {
                                'connector': 0,
                                'criteria': {
                                    'field': _[0],
                                    'values': [_[1]],
                                    "dataOperator" : int(_[2])}}
                        else:
                            filter_dict = {
                                'criteria': {
                                    'field': _[0],
                                    'values': [_[1]]}}
                        request_json['queries'][ind_]['whereClause'].append(filter_dict)
        request_json['options']['fetchFileProperties'] = True
        return request_json

    def azuread_browse_options_builder(self, options):
        """ build browse options
            Args:
                options    (dict)    browse option from impoort
            Return:
                options    (list)    create formated options based on import
            Raise:
                None
        """
        if "filters" not in options:
            options['filters'] = [("76", "00000000000000000000000000000001", "9"),
                                  ("76", "00000000000000000000000000000001", "9")]
        if "operation" not in options:
            options['operation'] = "browse"
            options['page_size'] = 20
            options['skip_node'] = 0
        return options



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""Main file for performing alert operations.

Alerts and Alert are 2 classes defined in this file.

Alerts: Class for representing all the Alerts

Alert: Class for a single alert selected


Alerts:
    __init__(commcell_object)   --  initialise object of Alerts class associated with
    the specified commcell

    __str__()                   --  returns all the alerts associated with the commcell

    __repr__()                  --  returns the string for the instance of the Alerts class

    __len__()                   --  returns the number of alerts configured on the Commcell

    __getitem__()               --  returns the name of the alert for the given alert Id or the
    details for the given alert name

    _get_alerts()               --  gets all the alerts associated with the commcell specified

    _get_entities()             --  returns the list of associations for an alert

    _get_alert_json()           --  returns the dict/json required to create an alert

    get_alert_sender()          -- returns the mail sender name as set in the Email server

    create_alert(alert_name)    --  returns the instance of Alert class for created alert

    has_alert(alert_name)       --  checks whether the alert exists or not

    get(alert_name)             --  returns the alert class object of the input alert name

    delete(alert_name)          --  removes the alerts from the commcell of the specified alert

    console_alerts()            --  returns the list of all console alerts

    console_alert()             -- returns console alert details for a console alert with given livefeedid

    refresh()                   --   refresh the alerts associated with the commcell

Alerts Attributes
------------------
    **all_alerts**              --  returns the dict of all the alerts associated
    with the commcell and their information such as name, id and category


Alert:
    __init__(commcell_object,
             alert_name,
             alert_id=None)       --  initialise object of alert with the specified commcell name
    and id, and associated to the specified commcell

    __repr__()                    --  return the alert name with description and category,
    the alert is associated with

    _get_alert_id()               --  method to get the alert id, if not specified in __init__

    _get_alert_properties()       --  get the properties of this alert

    _get_alert_category()         --  return the category of the alert

    _modify_alert_properties      --  modifies the alert properties

    alert_name(name)              --  sets the alert name

    alert_severity(severity)      --  sets the alert severity

    notification_types(list)      --  sets the notifications types

    entities(entities_dict)       --  sets the entities/associations

    enable()                      --  enables the alert

    disable()                     --  disables the alert

    enable_notification_type()    --  enables notification type of alert

    disable_notification_type()   --  disables notification type of alert

    refresh()                     --  refresh the properties of the Alert

Alert Attributes
------------------
    **alert_id**                --  returns the id of an alert

    **alert_name**              --  gets the name of an alert

    **alert_type**              --  returns the type of an alert

    **alert_category**          --  returns the category of an alert

    **alert_severity**          --  returns the severity of an alert

    **alert_criteria**          --  returns the criteria of an alert

    **notification_types**      --  returns the notification types of an alert

    **description**             --  returns the description of an alert

    **users_list**              --  returns the list of users associated with the alert

    **user_group_list**         --  returns the list of user groups associated with the alert

    **entities**                --  returns the list of entities associated with an alert

    **email_recipients**        --  returns the list of email recipients associated to the alert
"""

from __future__ import absolute_import
from __future__ import unicode_literals
import xml.etree.ElementTree as ET
from .exception import SDKException


class Alerts(object):
    """Class for getting all the Alerts associated with the commcell."""

    def __init__(self, commcell_object):
        """Initialize object of the Alerts class.

            Args:
                commcell_object (object)  --  instance of the Commcell class

            Returns:
                object - instance of the Alerts class
        """
        self._commcell_object = commcell_object
        self._ALERTS = commcell_object._services['GET_ALL_ALERTS']
        self._cvpysdk_object = commcell_object._cvpysdk_object
        self._services = commcell_object._services
        self._update_response_ = commcell_object._update_response_
        self._alerts = None

        self._notification_types = {
            'email': 1,
            'snmp': 4,
            'event viewer': 8,
            'save to disk': 512,
            'rss feeds': 1024,
            'console alerts': 8192,
            'scom': 32768,
            'workflow': 65536,
            'content indexing': 131072
        }
        self.refresh()

    def __str__(self):
        """Representation string consisting of all alerts of the Commcell.

            Returns:
                str - string of all the alerts for a commcell
        """
        representation_string = "{:^5}\t{:^50}\t{:^80}\t{:^30}\n\n".format(
            'S. No.', 'Alert', 'Description', 'Category'
        )

        for index, alert_name in enumerate(self._alerts):
            alert_description = self._alerts[alert_name]['description']
            alert_category = self._alerts[alert_name]['category']
            sub_str = '{:^5}\t{:50}\t{:80}\t{:30}\n'.format(
                index + 1,
                alert_name,
                alert_description,
                alert_category
            )
            representation_string += sub_str

        return representation_string.strip()

    def __repr__(self):
        """Representation string for the instance of the Alerts class."""
        return "Alerts class instance for Commcell: '{0}'".format(
            self._commcell_object.commserv_name
        )

    def __len__(self):
        """Returns the number of the alerts configured on the Commcell."""
        return len(self.all_alerts)

    def __getitem__(self, value):
        """Returns the name of the alert for the given alert ID or
            the details of the alert for given alert Name.

            Args:
                value   (str / int)     --  Name or ID of the alert

            Returns:
                str     -   name of the alert, if the alert id was given

                dict    -   dict of details of the alert, if alert name was given

            Raises:
                IndexError:
                    no alert exists with the given Name / Id

        """
        value = str(value)

        if value in self.all_alerts:
            return self.all_alerts[value]
        else:
            try:
                return list(filter(lambda x: x[1]['id'] == value, self.all_alerts.items()))[0][0]
            except IndexError:
                raise IndexError('No alert exists with the given Name / Id')

    def _get_alerts(self):
        """Gets all the alerts associated with the commcell

            Returns:
                dict - consists of all alerts of the commcell
                    {
                         "alert1_name": {
                             "id": alert1_id,
                             "category": alert1_category
                         },
                         "alert2_name": {
                             "id": alert2_id,
                             "category": alert2_category
                         }
                    }

            Raises:
                SDKException:
                    if response is empty

                    if response is not success
        """
        flag, response = self._cvpysdk_object.make_request('GET', self._ALERTS)

        if flag:
            if response.json() and 'alertList' in response.json():
                alerts_dict = {}

                for dictionary in response.json()['alertList']:
                    temp_dict = {}

                    temp_name = dictionary['alert']['name'].lower()
                    temp_id = str(dictionary['alert']['id']).lower()
                    temp_description = dictionary['description'].lower()
                    temp_category = dictionary['alertCategory']['name'].lower()

                    temp_dict['id'] = temp_id
                    temp_dict['description'] = temp_description
                    temp_dict['category'] = temp_category

                    alerts_dict[temp_name] = temp_dict

                    del temp_dict

                return alerts_dict
            else:
                raise SDKException('Response', '102')
        else:
            response_string = self._update_response_(response.text)
            raise SDKException('Response', '101', response_string)


    def _get_entities(self, entities):
        """Returns the list of entities associations for an alert

        Args:
            entities    (dict)  --  dictionary of entities for an alert

        Raise:
            SDKException:
                if entities is not an instance of dictionary

        Returns:
            list  -  a list of associations for an alert

        """
        if not isinstance(entities, dict):
            raise SDKException('Alert', '101')

        # policies not handled as

        entity_dict = {
            "clients": {
                "clientName": "client_name",
                "clientId": "client_id",
                "_type_": 3
            },
            "client_groups": {
                "clientGroupName": "clientgroup_name",
                "clientGroupId": "clientgroup_id",
                "_type_": 28
            },
            "users": {
                "userName": "user_name",
                "userId": "user_id",
                "_type_": 13
            },
            "user_groups": {
                "userGroupName": "user_group_name",
                "userGroupId": "user_group_id",
                "_type_": 15
            },
            "disk_libraries": {
                "libraryName": "library_name",
                "libraryId": "library_id",
                "_type_": 9
            },
            "media_agents": {
                "mediaAgentName": "media_agent_name",
                "mediaAgentId": "media_agent_id",
                "_type_": 11
            },
            "storage_policies": {
                "storagePolicyName": "storage_policy_name",
                "storagePolicyId": "storage_policy_id",
                "_type_": 17
            }
        }

        associations = []

        for entity, values in entities.items():
            if entity == "entity_type_names":
                for value in values:
                    if value == "ALL_CLIENT_GROUPS_ENTITY":
                        entity_type = 27
                    else:
                        entity_type = 2
                    temp_dict = {
                        "entityTypeName": value,
                        "_type_": entity_type
                    }
                    associations.append(temp_dict)
            else:
                entity_obj = getattr(self._commcell_object, entity)

                # this will allows us to loop through even for single item
                values = values.split() if not isinstance(values, list) else values

                for value in values:
                    temp_dict = entity_dict[entity].copy()
                    for name, entity_attr in temp_dict.items():
                        if name != "_type_":
                            try: # to convert the string values to int types
                                temp_dict[name] = int(getattr(entity_obj.get(value), entity_attr))
                            except ValueError:
                                temp_dict[name] = getattr(entity_obj.get(value), entity_attr)
                    associations.append(temp_dict)

        return associations


    def _get_alert_json(self, alert_json):
        """To form the json required to create an alert

        Args:
            alert_json    (dict)  --  a dictionary to create an alert

        Returns:
            dict  -  a constructed dictionary needed to create an alert
        """
        alert_detail = {
            "alertDetail": {
                "alertType": alert_json.get("alert_type"),
                "notifType": [n_type for n_type in alert_json.get("notif_type", [8192])],
                "notifTypeListOperationType": alert_json.get("notifTypeListOperationType", 0),
                "alertSeverity": alert_json.get("alertSeverity", 0),
                "EscnonGalaxyUserList":{
                    "nonGalaxyUserOperationType": alert_json.get("nonGalaxyUserOperationType", 0)
                },
                "locale":{
                    "localeID":alert_json.get("localeID", 0)
                },
                "EscUserList":{
                    "userListOperationType":alert_json.get("userListOperationType", 0)
                },
                "EscUserGroupList":{
                    "userGroupListOperationType": alert_json.get("userGroupListOperationType", 0)
                },
                "alertrule":{
                    "alertName": alert_json.get("alert_name")
                },
                "criteria":{
                    "criteria": alert_json.get("criteria")
                },
                "userList":{
                    "userListOperationType":alert_json.get("userListOperationType", 0),
                    "userList":[{"userName":user} for user in alert_json.get("users", ["admin"])]
                },
                "EntityList":{
                    "associationsOperationType":alert_json.get("associationsOperationType", 0),
                    "associations": self._get_entities(alert_json.get("entities", dict()))
                }
            }
        }

        # Check if paramsList is present or not
        if alert_json.get("paramsList"):
            alert_detail["alertDetail"]["criteria"]["paramsList"] = alert_json.get("paramsList")

        # Check if additonal mail recipents exist
        if alert_json.get("nonGalaxyList"):
            alert_detail["alertDetail"]["nonGalaxyList"] = alert_json.get("nonGalaxyList")

        if alert_json.get("user_groups"):
            alert_detail["alertDetail"]["userGroupList"] = {
                "userGroupListOperationType":alert_json.get("userGroupListOperationType", 0),
                "userGroupList":[
                    {
                        "userGroupName":user_grp
                    } for user_grp in alert_json.get("user_groups")
                ]
            }

        return alert_detail

    def get_alert_sender(self):
        """
            Returns the Alert Sender name
        """
        get_alert = self._services['EMAIL_SERVER']
        flag, response = self._cvpysdk_object.make_request('GET', get_alert)
        if flag:
            if response.json():
                sender = response.json()["senderInfo"]['senderName']
                if not sender:
                    sender = response.json()["senderInfo"]['senderAddress']
                return sender
            else:
                raise SDKException('Alert', '102', "Failed to get sender address")
        else:
            response_string = self._update_response_(response.text)
            raise SDKException('Response', '101', response_string)


    def create_alert(self, alert_dict):
        """Creates a new Alert for CommCell

        Args:
            alert_dict    (dict)  --  dictionary required to create an alert

        Returns:
            object  -  instance of the Alert class for this new alert

        Raises:
            SDKException:
                if input argument is not an instance of dict

                if alert with given name already exists

                if failed to create an alert

                if response is not success

                if response is empty
        """
        if not isinstance(alert_dict, dict):
            raise SDKException('Alert', '101')

        # required alert json
        alert_json = self._get_alert_json(alert_dict)
        alert_name = alert_json["alertDetail"]["alertrule"]["alertName"]
        if self.has_alert(alert_name):
            raise SDKException('Alert', '102', 'Alert "{0}" already exists.'.
                               format(alert_name))

        post_alert = self._services['GET_ALL_ALERTS']
        flag, response = self._cvpysdk_object.make_request(
            'POST', post_alert, alert_json)

        if flag:
            if response.json():
                error_dict = response.json()["errorResp"]
                error_code = str(error_dict["errorCode"])

                if error_code == "0":
                    self.refresh()
                    return self.get(alert_name)
                else:
                    error_message = ""

                    if 'errorMessage' in error_dict:
                        error_message = error_dict['errorMessage']

                    if error_message:
                        raise SDKException(
                            'Alert', '102', 'Failed to create Alert\nError: "{}"'.format(
                                error_message
                            )
                        )
                    else:
                        raise SDKException(
                            'Alert', '102', "Failed to create Alert")
            else:
                raise SDKException('Response', '102')
        else:
            response_string = self._update_response_(response.text)
            raise SDKException('Response', '101', response_string)


    @property
    def all_alerts(self):
        """Returns the dict of all the alerts configured on this commcell

            dict - consists of all alerts of the commcell
                    {
                         "alert1_name": {
                             "id": alert1_id,
                             "category": alert1_category
                         },
                         "alert2_name": {
                             "id": alert2_id,
                             "category": alert2_category
                         }
                    }
        """
        return self._alerts

    def has_alert(self, alert_name):
        """Checks if a alert exists for the commcell with the input alert name.

            Args:
                alert_name (str)  --  name of the alert

            Returns:
                bool - boolean output whether the alert exists for the commcell or not

            Raises:
                SDKException:
                    if type of the alert name argument is not string
        """
        if not isinstance(alert_name, str):
            raise SDKException('Alert', '101')

        return self._alerts and alert_name.lower() in self._alerts

    def get(self, alert_name):
        """Returns a alert object of the specified alert name.

            Args:
                alert_name (str)  --  name of the alert

            Returns:
                object - instance of the Alert class for the given alert name

            Raises:
                SDKException:
                    if type of the alert name argument is not string

                    if no alert exists with the given name
        """
        if not isinstance(alert_name, str):
            raise SDKException('Alert', '101')
        else:
            alert_name = alert_name.lower()

            if self.has_alert(alert_name):
                return Alert(
                    self._commcell_object, alert_name,
                    self._alerts[alert_name]['id'],
                    self._alerts[alert_name]['category']
                )

            raise SDKException('Alert', '102', 'No Alert exists with name: {0}'.format(alert_name))

    def console_alerts(self, page_number=1, page_count=1):
        """Returns the console alerts from page_number to the number of pages asked for page_count

            Args:
                page_number (int)  --  page number to get the alerts from

                page_count  (int)  --  number of pages to get the alerts of

            Raises:
                SDKException:
                    if type of the page number and page count argument is not int

                    if response is empty

                    if response is not success

            Returns:
                str - String representation of console alerts if version is less than SP23
                object - json response object for console alerts if version greater than or equal to SP23
        """
        if not (isinstance(page_number, int) and isinstance(page_count, int)):
            raise SDKException('Alert', '101')

        console_alerts = self._services['GET_ALL_CONSOLE_ALERTS'] % (
            page_number, page_count)

        flag, response = self._cvpysdk_object.make_request('GET', console_alerts)

        if flag:
            if response.json() and 'totalNoOfAlerts' in response.json():
                if self._commcell_object.commserv_version >= 23:
                    return response.json()

                o_str = "Total Console Alerts found: {0}".format(
                    response.json()['totalNoOfAlerts']
                )

                o_str += "\n{:^5}\t{:^50}\t{:^50}\t{:^50}\n\n".format(
                    'S. No.', 'Alert', 'Type', 'Criteria'
                )

                for index, dictionary in enumerate(response.json()['feedsList']):
                    o_str += '{:^5}\t{:50}\t{:^50}\t{:^50}\n'.format(
                        index + 1,
                        dictionary['alertName'],
                        dictionary['alertType'],
                        dictionary['alertcriteria']
                    )

                return o_str
            else:
                raise SDKException('Response', '102')
        else:
            response_string = self._update_response_(response.text)
            raise SDKException('Response', '101', response_string)

    def console_alert(self, live_feed_id):
        """Returns the console console alert with given live_feed_id

            Args:
                live_feed_id (int)  --  Live feed ID of console alert to fetch

            Raises:
                SDKException:
                    if type of the live_feed_id argument is not int

                    if response is empty

                    if response is not success

            Returns:
                object - Console alert json object for given live_feed_id
        """
        if not (isinstance(live_feed_id, int)):
            raise SDKException('Alert', '101')

        console_alerts = self._services['GET_CONSOLE_ALERT'] % (
            live_feed_id)

        flag, response = self._cvpysdk_object.make_request('GET', console_alerts)

        if flag:
            if response and response.json() and 'description' in response.json():
                return response.json()
            else:
                raise SDKException('Response', '102')
        else:
            response_string = self._update_response_(response.text)
            raise SDKException('Response', '101', response_string)

    def delete(self, alert_name):
        """Deletes the alert from the commcell.

            Args:
                alert_name (str)  --  name of the alert

            Raises:
                SDKException:
                    if type of the alert name argument is not string

                    if failed to delete the alert

                    if no alert exists with the given name
        """
        if not isinstance(alert_name, str):
            raise SDKException('Alert', '101')
        alert_name = alert_name.lower()

        if self.has_alert(alert_name):
            alert_id = self._alerts[alert_name]['id']
            alert = self._services['ALERT'] % (alert_id)

            flag, response = self._cvpysdk_object.make_request(
                'DELETE', alert
            )

            if flag:
                if response.json():
                    if 'errorCode' in response.json():
                        if response.json()['errorCode'] == 0:
                            # initialize the alerts again
                            # to refresh with the latest alerts
                            self.refresh()
                        else:
                            raise SDKException('Alert', '102', response.json()['errorMessage'])
                else:
                    raise SDKException('Response', '102')
            else:
                response_string = self._update_response_(response.text)
                exception_message = 'Failed to delete alert\nError: "{0}"'.format(
                    response_string
                )

                raise SDKException('Alert', '102', exception_message)
        else:
            raise SDKException(
                'Alert', '102', 'No alert exists with name: {0}'.format(alert_name)
            )

    def refresh(self):
        """Refresh the alerts associated with the Commcell."""
        self._alerts = self._get_alerts()


class Alert(object):
    """Class for performing operations for a specific alert."""

    def __init__(self, commcell_object, alert_name, alert_id=None, alert_category=None):
        """Initialise the Alert class instance.

            Args:
                commcell_object (object)  --  instance of the Commcell class

                alert_name      (str)     --  name of the alert

                alert_id        (str)     --  id of the alert
                    default: None

                alert_category  (str)     --  name of the alert category
                    default: None

            Returns:
                object - instance of the ALert class
        """

        self._commcell_object = commcell_object
        self._cvpysdk_object = commcell_object._cvpysdk_object
        self._services = commcell_object._services
        self._update_response_ = commcell_object._update_response_

        self._alerts_obj = Alerts(self._commcell_object)
        self._alert_name = alert_name.lower()
        self._alert_detail = None
        if alert_id:
            self._alert_id = str(alert_id)
        else:
            self._alert_id = self._get_alert_id()

        if alert_category:
            self._alert_category = alert_category
        else:
            self._alert_category = self._get_alert_category()

        self._ALERT = self._services['ALERT'] % (self.alert_id)
        self._all_notification_types = {
            'email': 1,
            'snmp': 4,
            'event viewer': 8,
            'save to disk': 512,
            'rss feeds': 1024,
            'console alerts': 8192,
            'scom': 32768,
            'workflow': 65536,
            'content indexing': 131072
        }

        self._alert_severity = None
        self._alert_type = None
        self._alert_type_id = None
        self._description = None
        self._criteria = []
        self._entities_list = []
        self._users_list = []
        self._user_group_list = []
        self._notification_types = []

        self.refresh()

    def __repr__(self):
        """String representation of the instance of this class."""
        representation_string = 'Alert class instance for Alert: "{0}", Alert Type: "{1}"'
        return representation_string.format(self.alert_name, self._alert_type)

    def _get_alert_id(self):
        """Gets the alert id associated with this alert.

            Returns:
                str - id associated with this alert
        """
        return self._alerts_obj.get(self.alert_name).alert_id

    def _get_alert_category(self):
        """Gets the alert category associated with this alert.

            Returns:
                str - alert category name associated with this alert
        """
        return self._alerts_obj.get(self.alert_name).alert_category

    def _get_alert_properties(self):
        """Gets the alert properties of this alert.

            Returns:
                dict - dictionary consisting of the properties of this alert

            Raises:
                SDKException:
                    if response is empty

                    if response is not success
        """
        flag, response = self._cvpysdk_object.make_request('GET', self._ALERT)

        if flag:
            if response.json() and 'alertDetail' in response.json().keys():
                self._alert_detail = response.json()['alertDetail']
                if 'alertSeverity' in self._alert_detail:
                    self._alert_severity = self._alert_detail['alertSeverity']

                if 'criteria' in self._alert_detail:
                    criterias = self._alert_detail['criteria']
                    for criteria in criterias:
                        self._criteria.append({
                            'criteria_value': criteria['value'] if 'value' in criteria else None,
                            'criteria_id':
                            str(criteria['criteriaId']) if 'criteriaId' in criteria else None,
                            'esclation_level':
                            criteria['esclationLevel'] if 'esclationLevel' in criteria else None
                        })

                if 'alert' in self._alert_detail:
                    alert = self._alert_detail['alert']

                    if 'description' in alert:
                        self._description = alert['description']

                    if 'alertType' in alert and 'name' in alert['alertType']:
                        self._alert_type = alert['alertType']['name']
                        self._alert_type_id = alert['alertType']['id']

                if 'xmlEntityList' in self._alert_detail:
                    entity_xml = ET.fromstring(self._alert_detail['xmlEntityList'])
                    self._entities_list = []
                    for entity in entity_xml.findall("associations"):
                        if entity.find("flags") is not None:
                            if entity.find("flags").attrib["exclude"] != "1":
                                self._entities_list.append(entity.attrib)
                        else:
                            self._entities_list.append(entity.attrib)

                # to convert the ids to int type
                for entity in self._entities_list:
                    for key, value in entity.items():
                        try:
                            entity[key] = int(value) # to change the ids to type int
                        except ValueError:
                            pass

                if 'regularNotifications' in self._alert_detail:
                    self._notification_types = self._alert_detail["regularNotifications"]

                if 'userList' in self._alert_detail:
                    self._users_list = [user['name'] for user in self._alert_detail['userList']]
                else:
                    self._users_list = []

                if 'userGroupList' in self._alert_detail:
                    self._user_group_list = [grp['name'] for grp in self._alert_detail['userGroupList']]
                else:
                    self._user_group_list = []

                self._email_recipients = []
                if 'nonGalaxyUserList' in self._alert_detail:
                    self._email_recipients = [email['name'] for email in self._alert_detail['nonGalaxyUserList']]

            else:
                raise SDKException('Response', '102')
        else:
            response_string = self._update_response_(response.text)
            raise SDKException('Response', '101', response_string)

    def _modify_alert_properties(self):
        """
        modifies the properties of an alert
        Exception:
            if modification of the alert failed
        """

        request_json = {
            "alertDetail":{
                "alertDetail": {
                    "alertType": self._alert_type_id, # this should not be changed
                    "notifType": self._notification_types,
                    "alertSeverity": self._alert_severity,
                    "alertrule": {
                        "alertName": self._alert_name
                    },
                    "criteria": {
                        "criteria": int(self._criteria[0]['criteria_id'])
                    },
                    "userList": {
                        "userListOperationType": 1,
                        "userList": [{"userName": user} for user in self._users_list]
                    },
                    "userGroupList": {
                        "userGroupListOperationType": 1,
                        "userGroupList": [{"userGroupName": user} for user in self._user_group_list]
                    },
                    "nonGalaxyList": {
                        "nonGalaxyUserList": [{"nonGalaxyUser": email} for email in self._email_recipients]
                    },
                    "EntityList": {
                        "associations": self._entities_list
                    }
                }
            }
        }

        modify_alert = self._services['MODIFY_ALERT'] % (self.alert_id)
        flag, response = self._cvpysdk_object.make_request(
            'POST', modify_alert, request_json
        )

        if flag:
            if response.json():
                error_code = str(response.json()['errorCode'])
                if error_code == '0':
                    self.refresh()
                    return
                else:
                    o_str = 'Failed to update properties of Alert\nError: "{0}"'
                    o_str = o_str.format(response.json()['errorMessage'])
                    raise SDKException('Alert', '102', o_str)
            else:
                raise SDKException('Response', '102')
        else:
            response_string = self._update_response_(response.text)
            raise SDKException('Response', '101', response_string)



    @property
    def name(self):
        """Returns the Alert display name """
        return self._alert_detail['alert']['alert']['name']

    @property
    def alert_name(self):
        """Treats the alert name as a read-only attribute."""
        return self._alert_name

    @alert_name.setter
    def alert_name(self, name):
        """Modifies the Alert name"""
        if not isinstance(name, str):
            raise SDKException('Alert', '101')

        self._alert_name = name
        self._modify_alert_properties()

    @property
    def alert_id(self):
        """Treats the alert id as a read-only attribute."""
        return self._alert_id

    @property
    def alert_type(self):
        """Treats the alert type as a read-only attribute."""
        return self._alert_type

    @property
    def alert_category(self):
        """Treats the alert category type id as a read-only attribute. """
        return self._alert_category.title()

    @property
    def alert_severity(self):
        """Treats the alert severity type id as a read-only attribute. """
        return self._alert_severity

    @alert_severity.setter
    def alert_severity(self, severity):
        """Modifies the Alert severity"""
        if not isinstance(severity, int):
            raise SDKException('Alert', '101')

        self._alert_severity = severity
        self._modify_alert_properties()

    @property
    def alert_criteria(self):
        """Treats the alert criteria as a read-only attribute."""
        return "\n".join([criteria["criteria_value"] for criteria in self._criteria])

    @property
    def notification_types(self):
        """Treats the alert notif types as a read-only attribute."""
        notif_types = []
        for notif, notif_id in self._all_notification_types.items():
            if notif_id in self._notification_types:
                notif_types.append((notif_id, notif.title()))
        return notif_types

    @notification_types.setter
    def notification_types(self, notif_types):
        """Treats the alert notif types as a read-only attribute."""
        if not isinstance(notif_types, list):
            raise SDKException('Alert', '102')
        try:
            ntypes = [self._all_notification_types[ntype.lower()] for ntype in notif_types]
        except KeyError as notif_type:
            raise SDKException(
                'Alert',
                '102',
                'No notification type with name {0} exists'.format(notif_type)
            )
        self._notification_types = ntypes
        self._modify_alert_properties()

    @property
    def entities(self):
        """Treats the alert associations as a read-only attribute. """
        return self._entities_list

    @entities.setter
    def entities(self, entity_json):
        """Modifies the Alert entities"""
        if not isinstance(entity_json, dict):
            raise SDKException('Alert', '101')

        self._entities_list = self._alerts_obj._get_entities(entity_json)
        self._modify_alert_properties()

    @property
    def email_recipients(self):
        """returns the email recipients associated to the alert"""
        return self._email_recipients

    @email_recipients.setter
    def email_recipients(self, email_recipients):
        """Modifies the email_recipients for the alert"""
        self._email_recipients.extend(email_recipients)
        self._modify_alert_properties()

    @property
    def description(self):
        """Treats the alert description as a read-only attribute."""
        return self._description

    @description.setter
    def description(self, description):
        """Modifies the Alert description"""
        self._description = description
        self._modify_alert_properties()
    
    @property
    def users_list(self):
        """Treats the users list as a read-only attribute."""
        return self._users_list

    @users_list.setter
    def users_list(self, users_list):
        """Modifies the users list"""
        self._users_list = users_list
        self._modify_alert_properties()
    
    @property
    def user_group_list(self):
        """Treats the user group list as a read-only attribute."""
        return self._user_group_list

    @user_group_list.setter
    def user_group_list(self, user_group_list):
        """Modifies the user group list"""
        self._user_group_list = user_group_list
        self._modify_alert_properties()

    def enable_notification_type(self, alert_notification_type):
        """Enable the notification type.

            Args:
                alert_notification_type (str)  --  alert notification to enable

            Raises:
                SDKException:
                    if type of alert notification argument is not string

                    if failed to enable notification type

                    if response is empty

                    if response is not success

                    if no notification type exists with the name provided
        """
        if not isinstance(alert_notification_type, str):
            raise SDKException('Alert', '101')

        if alert_notification_type.lower() in self._all_notification_types:
            alert_notification_type_id = self._all_notification_types[
                alert_notification_type.lower()]

            enable_request = self._services['ENABLE_ALERT_NOTIFICATION'] % (
                self.alert_id, alert_notification_type_id
            )

            flag, response = self._cvpysdk_object.make_request(
                'POST', enable_request
            )

            if flag:
                if response.json():
                    error_code = str(response.json()['errorCode'])
                    if error_code == '0':
                        return
                    else:
                        raise SDKException('Alert', '102', response.json()['errorMessage'])
                else:
                    raise SDKException('Response', '102')
            else:
                response_string = self._update_response_(response.text)
                raise SDKException('Response', '101', response_string)
        else:
            raise SDKException(
                'Alert',
                '102',
                'No notification type with name {0} exists'.format(alert_notification_type)
            )

    def disable_notification_type(self, alert_notification_type):
        """Disable the notification type.

            Args:
                alert_notification_type (str)  --  alert notification to disable

            Raises:
                SDKException:
                    if type of alert notification argument is not string

                    if failed to disable notification type

                    if response is empty

                    if response is not success

                    if no notification type exists with the name provided
        """
        if not isinstance(alert_notification_type, str):
            raise SDKException('Alert', '101')

        if alert_notification_type.lower() in self._all_notification_types:
            alert_notification_type_id = self._all_notification_types[
                alert_notification_type.lower()]

            disable_request = self._services['DISABLE_ALERT_NOTIFICATION'] % (
                self.alert_id, alert_notification_type_id
            )

            flag, response = self._cvpysdk_object.make_request(
                'POST', disable_request
            )

            if flag:
                if response.json():
                    error_code = str(response.json()['errorCode'])
                    if error_code == '0':
                        return
                    else:
                        raise SDKException('Alert', '102', response.json()['errorMessage'])
                else:
                    raise SDKException('Response', '102')
            else:
                response_string = self._update_response_(response.text)
                raise SDKException('Response', '101', response_string)
        else:
            raise SDKException(
                'Alert',
                '102',
                'No notification type with name {0} exists'.format(alert_notification_type)
            )

    def enable(self):
        """Enable an alert.

            Raises:
                SDKException:
                    if failed to enable alert

                    if response is empty

                    if response is not success
        """
        enable_request = self._services['ENABLE_ALERT'] % (self.alert_id)

        flag, response = self._cvpysdk_object.make_request('POST', enable_request)

        if flag:
            if response.json():
                error_code = str(response.json()['errorCode'])

                if error_code == "0":
                    return
                else:
                    error_message = ""

                    if 'errorMessage' in response.json():
                        error_message = response.json()['errorMessage']

                    if error_message:
                        raise SDKException(
                            'Alert', '102', 'Failed to enable Alert\nError: "{0}"'.format(
                                error_message
                            )
                        )
                    else:
                        raise SDKException('Alert', '102', "Failed to enable Alert")
            else:
                raise SDKException('Response', '102')
        else:
            response_string = self._update_response_(response.text)
            raise SDKException('Response', '101', response_string)

    def disable(self):
        """Disable an alert.

            Raises:
                SDKException:
                    if failed to disable alert

                    if response is empty

                    if response is not success
        """
        disable_request = self._services['DISABLE_ALERT'] % (self.alert_id)

        flag, response = self._cvpysdk_object.make_request(
            'POST', disable_request
        )

        if flag:
            if response.json():
                error_code = str(response.json()['errorCode'])

                if error_code == "0":
                    return
                else:
                    error_message = ""

                    if 'errorMessage' in response.json():
                        error_message = response.json()['errorMessage']

                    if error_message:
                        raise SDKException(
                            'Alert', '102', 'Failed to disable Alert\nError: "{0}"'.format(
                                error_message
                            )
                        )
                    else:
                        raise SDKException('Alert', '102', "Failed to disable Alert")
            else:
                raise SDKException('Response', '102')
        else:
            response_string = self._update_response_(response.text)
            raise SDKException('Response', '101', response_string)

    def refresh(self):
        """Refresh the properties of the Alert."""
        self._get_alert_properties()



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""" Main file for performing the software cache configuration related operations

CommserveCache   --  Class for performing operations on the CS cache
RemoteCache      --  Class for performing operations on the remote cache

CommServeCache
==============

    __init__(commcell_object)             --  initialize commcell_object of CommServeCache class
    associated with the commcell

    get_request_xml()                     --  returns request xml for cache and remote cache related operations

    get_cs_cache_path()                   --  returns CS cache path

    delete_cache()                        --  deletes CS cache

    commit_cache()                        --  commits CS cache

RemoteCache
==============

    __init__(commcell, client_object)     --  initialize commcell and client_object of RemoteCache class
    associated with the client

    get_remote_cache_path()               --  returns remote cache path, if exists, else None

    configure_remote_cache()              --  Configures client as remote cache

    configure_packages_to_sync()          --  Configures packages to sync for the remote cache

    assoc_entity_to_remote_cache()        --  Associates entity to the Remote Cache

    delete_remote_cache_contents()        --  deletes remote cache contents
"""
from xml.etree import ElementTree as ET
from ..exception import SDKException
from .deploymentconstants import UnixDownloadFeatures
from .deploymentconstants import WindowsDownloadFeatures
from .deploymentconstants import OSNameIDMapping


class CommServeCache(object):
    """"class for downloading software packages"""

    def __init__(self, commcell_object):
        """Initialize commcell_object of the Download class.

            Args:
                commcell_object (object)  --  instance of the Commcell class

            Returns:
                object - instance of the CommServeCache class

        """

        self.commcell_object = commcell_object
        self.request_xml = CommServeCache.get_request_xml()

    @staticmethod
    def get_request_xml():
        """Returns request xml for cache and remote cache related operations"""
        return """<EVGui_SetUpdateAgentInfoReq>
                <uaInfo deletePackageCache="" deleteUpdateCache="" swAgentOpType=""
                uaOpCode="0" uaPackageCacheStatus="0"
                 uaUpdateCacheStatus="0" >
                <uaName id="2" name=""/>
                <client _type_="3"/>
                </uaInfo>
                </EVGui_SetUpdateAgentInfoReq>
                """

    def get_cs_cache_path(self):
        """
        Returns CS cache path

        Returns:
            cs_cache_path (str) -- returns cs cache path

        Raises:
            SDKException:
            - Failed to execute the api

            - Response is incorrect/empty
        """
        try:
            response = self.commcell_object.get_gxglobalparam_value()
        except Exception:
            try:
                response = self.commcell_object.get_gxglobalparam_value('Patch Directory')
            except Exception:
                raise SDKException('Response', '101', 'Failed to execute api for get_cs_cache_path')
            if response == '':
                raise SDKException('Response', '102')
            return response
        if response['error']['errorCode'] != 0:
            error_message = "Failed with error: [{0}]".format(
                response['error']['errorMessage']
            )
            raise SDKException(
                'Response',
                '101',
                'Error Code:"{0}"\nError Message: "{1}"'.format(response['error']['errorCode'], error_message)
            )
        try:
            return response['commserveSoftwareCache']['storePatchlocation']
        except Exception:
            raise SDKException('Response', '102')

    def delete_cache(self):
        """
        Delete CS cache

        Raises:
            SDKException:
            - Failed to execute the api

            - Response is incorrect
        """
        root = ET.fromstring(self.request_xml)
        uaInfo = root.find(".//uaInfo")
        uaInfo.set('deletePackageCache', "1")
        uaInfo.set("deleteUpdateCache", "1")
        uaInfo.set("swAgentOpType", "1")

        response = self.commcell_object.qoperation_execute(ET.tostring(root))
        if response.get('errorCode') != 0:
            error_message = "Failed with error: [{0}]".format(
                response.get('errorMessage')
            )
            raise SDKException(
                'Response',
                '101',
                'Error Code:"{0}"\nError Message: "{1}"'.format(response.get('errorCode'), error_message)
            )

    def commit_cache(self):
        """
        Commits CS cache

        Raises:
            SDKException:
            - Failed to execute the api

            - Response is incorrect
        """

        root = ET.fromstring(self.request_xml)
        uaInfo = root.find(".//uaInfo")
        uaInfo.set('deletePackageCache', "0")
        uaInfo.set("deleteUpdateCache", "0")
        uaInfo.set("swAgentOpType", "4")

        response = self.commcell_object.qoperation_execute(ET.tostring(root))
        if response.get('errorCode') != 0:
            error_message = "Failed with error: [{0}]".format(
                response.get('errorMessage')
            )
            raise SDKException(
                'Response',
                '101',
                'Error Code:"{0}"\nError Message: "{1}"'.format(response.get('errorCode'), error_message)
            )


class RemoteCache(object):
    """"class for downloading software packages"""

    def __init__(self, commcell, client_name):
        """Initialize commcell_object of the Download class.

            Args:
                commcell (object)     --  commcell object
                client_name           --  client name

            Returns:
                object - instance of the RemoteCache class

        """
        self.commcell = commcell
        self.client_object = self.commcell.clients.get(client_name)
        self.request_xml = CommServeCache.get_request_xml()
        self._cvpysdk_object = commcell._cvpysdk_object
        self._services = commcell._services

    def get_remote_cache_path(self):
        """
        Returns remote cache path, if exists, else None

        Returns:
            remote_cache_path (str) - remote cache path of the client if exists
            None                    - otherwise

        Raises:
            SDKException:
            - Failed to execute the api

            - Response is incorrect/empty

        """
        request_xml = '<EVGui_GetUpdateAgentInfoReq />'
        response = self.commcell.qoperation_execute(request_xml)
        if response:
            try:
                for clients in response["uaInfo"]:
                    if clients['client']['clientName'] == self.client_object.client_name:
                        return clients["uaCachePath"]
                return None
            except Exception:
                raise SDKException('Response', '101')
        else:
            raise SDKException('Response', '102')

    def configure_remotecache(self, cache_path):
        """
        Configures client as remote cache

        Args:
              cache_path (str)  - Remote cache path

        Raises:
            SDKException:
            - Failed to execute the api

            - Response is incorrect
        """
        root = ET.fromstring(self.request_xml)
        uaInfo = root.find(".//uaInfo")
        uaInfo.set('uaCachePath', cache_path)
        uaInfo.set('uaOpCode', "5")
        uaInfo.attrib.pop("uaPackageCacheStatus")
        uaInfo.attrib.pop('uaUpdateCacheStatus')
        root.find("./uaInfo/uaName").set("id", self.client_object.client_id)
        root.find("./uaInfo/uaName").set("name", self.client_object.client_name)

        response = self.commcell.qoperation_execute(ET.tostring(root))
        if response.get('errorCode') != 0:
            error_message = "Failed with error: [{0}]".format(
                response.get('errorMessage')
            )
            raise SDKException(
                'Response',
                '101',
                'Error Code:"{0}"\nError Message: "{1}"'.format(response.get('errorCode'), error_message)
            )

    def configure_packages_to_sync(self, win_os=None, win_package_list=None, unix_os=None,
                                   unix_package_list=None):
        """
        Configures packages to sync for the remote cache

        Args:
            win_os 		(list)	 	-- list of windows oses to sync
            win_package_list  (list)-- list of windows packages to sync
            unix_os (list) 		  	-- list of unix oses to sync
            unix_package_list (list)-- list of unix packages to sync

        Raises:
            SDKException:
            - Failed to execute the api

            - Response is incorrect

            - Incorrect input

        Usage:
            commcell_obj.configure_packages_to_sync()

            win_os = ["WINDOWS_32", "WINDOWS_64"]
            unix_os = ["UNIX_LINUX64", "UNIX_AIX"]
            win_package_list = ["FILE_SYSTEM", "MEDIA_AGENT"]
            unix_package_list = ["FILE_SYSTEM", "MEDIA_AGENT"]

            OS_Name_ID_Mapping, WindowsDownloadFeatures and UnixDownloadFeatures enum is used for
            providing input to the configure_packages_to_sync method, it can be imported by

                >>> from cvpysdk.deployment.deploymentconstants import UnixDownloadFeatures
                    from cvpysdk.deployment.deploymentconstants import OS_Name_ID_Mapping
                    from cvpysdk.deployment.deploymentconstants import WindowsDownloadFeatures

        """
        if win_os:
            win_os_id = [eval(f"OSNameIDMapping.{each}.value") for each in win_os]
            win_packages = [eval(f"WindowsDownloadFeatures.{packages}.value") for packages in win_package_list]
        if unix_os:
            unix_os_id = [eval(f"OSNameIDMapping.{each}.value") for each in unix_os]
            unix_packages = [eval(f"UnixDownloadFeatures.{packages}.value") for packages in unix_package_list]

        if not win_os and not unix_os:
            qscript = f'''-sn QS_GranularConfigRemoteCache -si '{self.client_object.client_name}' -si SyncAll'''
        elif not unix_os:
            qscript = (f'''-sn QS_GranularConfigRemoteCache -si '{self.client_object.client_name}' -si SyncCustom '''
                       f'''-si {",".join(map(str, win_os_id))} -si {",".join(map(str, win_packages))}''')
        elif not win_os:
            qscript = (f'''-sn QS_GranularConfigRemoteCache -si '{self.client_object.client_name}' -si SyncCustom '''
                       f'''-si {",".join(map(str, unix_os_id))} -si {",".join(map(str, unix_packages))}''')
        else:
            qscript = (f'''-sn QS_GranularConfigRemoteCache -si '{self.client_object.client_name}' -si SyncCustom '''
                       f'''-si {",".join(map(str, win_os_id))} -si {",".join(map(str, win_packages))} '''
                       f'''-si {",".join(map(str, unix_os_id))} -si {",".join(map(str, unix_packages))}''')

        response = self.commcell._qoperation_execscript(qscript)
        if response.get('CVGui_GenericResp'):
            if response['CVGui_GenericResp']['@errorCode'] != 0:
                error_message = "Failed with error: [{0}]".format(
                    response['CVGui_GenericResp']['@errorMessage']
                )
            raise SDKException(
                'Response',
                '101',
                'Error Code:"{0}"\nError Message: "{1}"'.format(
                    response['CVGui_GenericResp']['@errorCode'],
                    error_message))

    def delete_remote_cache_contents(self):
        """
        Delete remote cache contents

        Raises:
            SDKException:
            - Failed to execute the api

            - Response is incorrect
        """
        root = ET.fromstring(self.request_xml)
        uaInfo = root.find(".//uaInfo")
        uaInfo.set('deletePackageCache', "1")
        uaInfo.set("deleteUpdateCache", "1")
        uaInfo.set("swAgentOpType", "1")
        root.find("./uaInfo/uaName").set("id", self.client_object.client_id)
        root.find("./uaInfo/uaName").set("name", self.client_object.client_name)

        response = self.commcell.qoperation_execute(ET.tostring(root))
        if response.get('errorCode') != 0:
            error_message = "Failed with error: [{0}]".format(
                response.get('errorMessage')
            )
            raise SDKException(
                'Response',
                '101',
                'Error Code:"{0}"\nError Message: "{1}"'.format(response.get('errorCode'), error_message)
            )

    def assoc_entity_to_remote_cache(self, client_name=None, client_group_name=None):
        """
            Points/Associates entity to the Remote Cache Client

                Args:
                    client_name (str)  -- The client which has to be pointed to Remote Cache

                    client_group_name (str)  -- The client_group which has to be pointed to Remote Cache

                Raises:
                    SDKException:
                    - Failed to execute the api

                    - Response is incorrect
        """

        if client_name is None and client_group_name is None:
            raise Exception("No clients or client groups to associate; Please provide a valid name")

        if client_name and client_name in self.commcell.clients.all_clients:
            entity_obj = self.commcell.clients.get(client_name)
            entity_id = entity_obj.client_id
            entity_name = entity_obj.client_name
            entity_type ="0"

        elif client_group_name in self.commcell.client_groups.all_clientgroups:
            entity_obj = self.commcell.client_groups.get(client_group_name)
            entity_id = entity_obj.clientgroup_id
            entity_name = entity_obj.clientgroup_name
            entity_type = "1"

        else:
            raise Exception("{0} does not exist".format(client_name if client_name else client_group_name))

        request_json = {
                "EVGui_SetUpdateAgentInfoReq" :{
                "uaInfo": {
                    "uaCachePath": self.get_remote_cache_path(),
                    "uaOpCode": "5",
                    "uaName": {
                        "id": self.client_object.client_id,
                        "name": self.client_object.client_name
                    }
                },
                "uaList": {
                    "addedList": {
                        "id": entity_id,
                        "name": entity_name,
                        "type": entity_type
                    }
                }
            }
        }

        flag, response = self._cvpysdk_object.make_request(
            'POST', self._services['EXECUTE_QCOMMAND'], request_json
        )

        if flag:
            if response.ok:
                if response.json():
                    if response.json().get('errorCode') != 0:
                        error_code = response.json().get('errorCode')
                        error_message = "Failed with error: [{0}]".format(
                            response.json().get('errorMessage')
                        )
                        raise SDKException(
                            'Response',
                            '101',
                            'Error Code:"{0}"\nError Message: "{1}"'.format(error_code, error_message)
                        )
                else:
                    raise SDKException('Response', '102')
            else:
                raise SDKException('Response', '102')
        else:
            raise SDKException('Response', '101')


# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""Main file for doing Name Change operations.

OperationType:  Class with the supported hostname change operations

NameChange: Class for doing operations for Name Change operations on clients and commcell.

NameChange:
    __init__(class_object)                          --  initialise object of the NameChange
                                                        class

    hostname()                                      --  gets the current hostname of the client or
                                                        commserver

    hostname(parameters_dict)                       --  sets the hostname from client or commserver
                                                        level

    display_name()                                  --  gets the display name of the client or
                                                        commserver

    display_name(display_name)                      --  sets the display name of the client or
                                                        commserver

    client_name()                                   --  gets the name of the client

    client_name(client_name)                        --  sets the name of the client

    domain_name()                                    --  gets the commserver hostname

    domain_name(domains_dict)                        --  sets the new domain name for the clients

    _client_name_change_op()                        --  performs client namechange based on the
                                                        setters

    _commcell_name_change_op(parameters_dict)       --  performs commserver namechange based on the
                                                        setters

    get_clients_for_name_change_post_ccm()          -- gets all the clients available for name change
                                                        post commcell migration

    name_change_post_ccm(parameters_dict)           -- perfoms name change for migrated clients post
                                                        commcell migration
"""

import re
from enum import Enum
from .exception import SDKException


class OperationType(Enum):
    """ Operation Types supported to get schedules of particular optype"""
    COMMSERVER_HOSTNAME_REMOTE_CLIENTS = 147
    COMMSERVER_HOSTNAME_AFTER_DR = 139
    CLIENT_HOSTNAME = "CLIENT_HOSTNAME"
    COMMSERVER_HOSTNAME = "COMMSERVER_HOSTNAME"


class NameChange(object):
    """Class for doing Name Change operations on clients and commcell"""

    def __init__(self, class_object):
        """Initializes an instance of the NameChange class to perform Name Change operations.

            Args:
                class_object (object)  --  instance of the client/commcell class

        """
        from .commcell import Commcell
        from .client import Client

        if isinstance(class_object, Commcell):
            self._commcell_object = class_object
            self._display_name = self._commcell_object.clients.get(self._commcell_object.
                                                                   commserv_hostname).display_name
            self._commcell_name = self._commcell_object.clients.get(self._commcell_object.
                                                                    commserv_hostname).commcell_name
            self._is_client = False

        elif isinstance(class_object, Client):
            self._client_object = class_object
            self._commcell_object = class_object._commcell_object
            self._client_hostname = self._client_object.client_hostname
            self._display_name = self._client_object.display_name
            self._client_name = self._client_object.client_name
            self._commcell_name = self._client_object.commcell_name
            self._new_name = None
            self._is_client = True

        self._cvpysdk_object = self._commcell_object._cvpysdk_object
        self._services = self._commcell_object._services
        self._update_response_ = self._commcell_object._update_response_

    @property
    def hostname(self):
        """
        Gets the client hostname or commserver hostname

         Returns:
                str - client hostname or commserver hostname
        """
        if self._is_client:
            return self._client_hostname
        else:
            return self._commcell_name

    @hostname.setter
    def hostname(self, parameters_dict):
        """
        Sets the client hostname or commserver hostname with the parameters provided
        Args:
            parameters_dict (str)      -- dictionary of parameters for namechange
                                    {
                                    "operation": Operation type to be performed on the client or
                                                commserver (OperationType)
                                    "ClientHostname":   Client hostname to be updated (str)
                                    "CommserverHostname":   Commserver hostname to be updated (str)
                                    "oldName":  old commserver hostname
                                    "newName":  new commserver hostname
                                    }

        """
        if self._is_client:
            if parameters_dict["operation"] == OperationType.CLIENT_HOSTNAME.value:
                if parameters_dict["ClientHostname"] is None:
                    raise SDKException('NameChange', '101')
                self._client_hostname = parameters_dict["ClientHostname"]
                self._client_name_change_op()
            if parameters_dict["operation"] == OperationType.COMMSERVER_HOSTNAME.value:
                if parameters_dict["CommserverHostname"] is None:
                    raise SDKException('NameChange', '102')
                self._commcell_name = parameters_dict["CommserverHostname"]
                self._client_name_change_op()
        else:
            if parameters_dict["operation"] == OperationType.COMMSERVER_HOSTNAME_REMOTE_CLIENTS.value:
                parameters_dict["oldName"] = self._commcell_name
                self._commcell_name_change_op(parameters_dict)
            elif parameters_dict["operation"] == OperationType.COMMSERVER_HOSTNAME_AFTER_DR.value:
                if parameters_dict["clientIds"] is None:
                    raise SDKException('NameChange', '105')
                parameters_dict["newName"] = self._commcell_name
                self._commcell_name_change_op(parameters_dict)

    @property
    def domain_name(self):
        """
        Gets the commserver hostname

        Returns:
                str - commserver hostname
        """
        return self._commcell_name

    @domain_name.setter
    def domain_name(self, domains_dict):
        """
        Sets the new domain name for the clients with the parameter provided
        Args:
            domains_dict (dict) -- new client domain name
                                    {
                                    "oldDomain": old client domain name (str)
                                    "newDomain": new client domain name (str)
                                    }

        """
        if domains_dict["oldDomain"] is None:
            raise SDKException('NameChange', '103')
        elif domains_dict["newDomain"] is None:
            raise SDKException('NameChange', '104')
        dict_domains = {
            "oldName": domains_dict["oldDomain"],
            "newName": domains_dict["newDomain"],
            "operation": 136
        }
        self._commcell_name_change_op(dict_domains)

    @property
    def display_name(self):
        """
        Gets the display name of the client or commserver

        Returns:
                str - client or commserver display name
        """
        if self._is_client:
            return self._display_name
        else:
            return self._display_name

    @display_name.setter
    def display_name(self, display_name):
        """
        Sets the display name of the client or commserver with the parameter provided
        Args:
            display_name (str) -- new client or commserver display name

        """
        if self._is_client:
            self._display_name = display_name
            self._client_name_change_op()
        else:
            dict_cs = {
                "oldName": self._display_name,
                "newName": display_name,
                "operation": 9811,
            }
            self._commcell_name_change_op(dict_cs)

    @property
    def client_name(self):
        """
        Gets the client name

        Returns:
                str - client name
        """
        if self._is_client:
            return self._client_name
        else:
            False

    @client_name.setter
    def client_name(self, client_name):
        """
        Sets the name of the client with the parameter provided
        Args:
            client_name (str) -- new client name

        """
        self._new_name = client_name
        self._client_name_change_op()

    def _client_name_change_op(self):
        """
        Performs the client namechange operations

            Raises:
            SDKException::
                if the client namechange failed

                if the response is empty
        """
        request_json = {
            "App_SetClientPropertiesRequest":
            {
                "clientProperties": {
                    "client": {
                        "displayName": self._display_name,
                        "clientEntity": {
                            "hostName": self._client_hostname,
                            "clientName": self._client_name,
                            "commCellName": self._commcell_name
                        }
                    }
                },
                "association": {
                    "entity": [
                        {
                            "clientName": self._client_name,
                            "newName": self._new_name
                        }
                    ]
                }
            }
        }
        flag, response = self._client_object._cvpysdk_object.make_request(
            'POST', self._services['EXECUTE_QCOMMAND'], request_json
        )

        if flag:
            if response.json():
                if 'errorMessage' in response.json():
                    # for errorMessage: "Operation Failed" errorCode: 7
                    # for errorMessage: "Error 0x911: Failed to process request due to invalid /
                    # entity information.Invalid clientId for clientName.\n"
                    # errorCode: 2 and others

                    error_message = "Failed to do namechange on client, " \
                                    "with errorCode [{0}], errorMessage [{1}]".format(
                                        response.json().get('errorCode'),
                                        response.json().get('errorMessage')
                                    )
                    raise SDKException('Client', '102', error_message)

                elif 'errorCode' in response.json().get('response')[0]:
                    error_code = str(
                        response.json().get('response')[0].get('errorCode'))
                    if error_code != '0':
                        error_message = "Failed to do namechange on client"
                        raise SDKException('Client', '102', error_message)
                else:
                    raise SDKException('Response', '102')
            else:
                raise SDKException(
                    'Response', '101', self._update_response_(
                        response.text))
        else:
            raise SDKException(
                'Response',
                '101',
                self._update_response_(
                    response.text))

    def _commcell_name_change_op(self, parameters_dict):
        """
        Performs the commcell namechange operations

        Args:
            parameters_dict (dict)          --  dictionary with common namechange parameters like
                                                old commserver hostname, new commserver hostname
                                                or old domain name, new domain name, client IDs.
                                                clientIds can be an empty list too.
                                                  {"newName",
                                                    "oldName",
                                                    "operation",
                                                    "clientIds"}

            Raises:
            SDKException::
                if the client namechange failed

                if the response is empty

        """

        request_json = {
            "EVGui_ClientNameControlReq":
            {
                "isPostMigration": "",
                "newName": parameters_dict.get("newName", ""),
                "destinationConfiguration": 0,
                "sourceConfiguration": 0,
                "setWithoutConditionFlag": 0,
                "oldName": parameters_dict.get("oldName", ""),
                "commCellId": 0,
                "operation": parameters_dict.get("operation", 0),
                "forceChangeName": 0,
                "clientList": parameters_dict.get("clientIds", [])

            }
        }
        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'POST', self._services['EXECUTE_QCOMMAND'], request_json
        )
        if flag:
            if response.json():
                if 'errorCode' in response.json().get('error'):
                    error_code = int(
                        response.json().get('error').get('errorCode'))
                    if error_code != 1:
                        # for errorString: "Failed to set Client properties."
                        # errorCode: 0 or others
                        error_message = "Failed to do namechange on commserver " \
                                        "with errorCode [{0}], errorString [{1}]".format(
                                            response.json().get('error').get('errorCode'),
                                            response.json().get('error').get('errorString')
                                        )
                        raise SDKException('Client', '102', error_message)

                    elif error_code == 1:
                        return True

                elif 'errorMessage' in response.json():
                        # for errorMessage: "Operation Failed" errorCode: 7
                    error_message = "Failed to do namechange on commserver " \
                                    "with errorCode [{0}], errorMessage [{1}]".format(
                                        response.json().get('errorCode'),
                                        response.json().get('errorMessage')
                                    )
                    raise SDKException('Client', '102', error_message)
                else:
                    raise SDKException('Response', '102')
            else:
                raise SDKException(
                    'Response', '101', self._update_response_(
                        response.text))
        else:
            raise SDKException(
                'Response',
                '101',
                self._update_response_(
                    response. text))

    def get_clients_for_name_change_post_ccm(self):
        """
            Gets clients available for name change after commcell migration.
            Raises:
            SDKException::
                if the client namechange failed
                if the response is empty
        """
        xml = """
            <EVGui_GetClientForNameControlReq>
            </EVGui_GetClientForNameControlReq>
        """
        flag, response = self._cvpysdk_object.make_request(
            'POST', self._services['EXECUTE_QCOMMAND'], xml
        )
        def get_clients(response):
            clients_list = []
            all_clients = response.json()["clientList"]
            for client in all_clients:
                temp_dict = {}
                name = client.get("name", "")
                domain = client.get("domain", "")
                cs_host_name = client.get("csHostName", "")
                if name + "." + domain != cs_host_name and name != cs_host_name:
                    clients_list.append({"csHostname": cs_host_name, "name": name})
            return clients_list
        if flag:
            if response.json():
                if 'errorCode' in response.json().get('error'):
                    error_code = int(
                        response.json().get('error').get('errorCode'))
                    if error_code != 1:
                        # for errorString: "Failed to get clients for name change operation"
                        # errorCode: 0 or others
                        error_message = "Failed to get clients for name change operation" \
                                        "with errorCode [{0}], errorString [{1}]".format(
                            response.json().get('error').get('errorCode'),
                            response.json().get('error').get('errorString')
                        )
                        raise SDKException('Client', '102', error_message)
                    elif error_code == 1:
                        return get_clients(response)
                elif 'errorMessage' in response.json():
                    error_message = "Failed to get clients for name change operation" \
                                    "with errorCode [{0}], errorMessage [{1}]".format(
                        response.json().get('errorCode'),
                        response.json().get('errorMessage')
                    )
                    raise SDKException('Client', '102', error_message)
                else:
                    raise SDKException('Response', '102')
            else:
                raise SDKException(
                    'Response', '101', self._update_response_(
                        response.text))
        else:
            raise SDKException(
                'Response',
                '101',
                self._update_response_(
                    response.text))

    def name_change_post_ccm(self, parameters_dict):
        """
        Performs the commcell namechange for clients post commcell migration
        Args:
            parameters_dict (dict)      --  contains old commcell hostname, new commcell hostname,
                                            Ids of clients on which name change is to be performed
                                            {
                                            "sourceCommcellHostname": "source-1"
                                            "destinationCommcellHostname": "dest-1"
                                            "clientIds": ["id1", "id2"]
                                            }
            Raises:
            SDKException::
                if the client namechange failed
                if the response is empty
        """
        name_change_xml = """
            <EVGui_ClientNameControlReq 
                commCellId="0" 
                destinationConfiguration="2" 
                isPostMigration="1" 
                newName="{0}"
                oldName="{1}"
                operation="139" 
                setWithoutConditionFlag="0" 
                sourceConfiguration="2"> 
                {2}
            </EVGui_ClientNameControlReq>
        """
        client_tag = """
            <clientList val= "{0}"/>
        """
        clients_string = ""
        for clients_id in parameters_dict.get("clientIds", []):
            clients_string += client_tag.format(clients_id)
        name_change_xml = name_change_xml.format(parameters_dict["destinationCommcellHostname"],
                                                 parameters_dict["sourceCommcellHostname"],
                                                 clients_string)
        flag, response = self._cvpysdk_object.make_request(
            'POST', self._services['EXECUTE_QCOMMAND'], name_change_xml
        )
        if flag:
            if response.json():
                if 'errorCode' in response.json().get('error'):
                    error_code = int(
                        response.json().get('error').get('errorCode'))
                    if error_code != 1:
                        error_message = "Failed to perform name change operation" \
                                        "with errorCode [{0}], errorString [{1}]".format(
                                        response.json().get('error').get('errorCode'),
                                        response.json().get('error').get('errorString')
                                        )
                        raise SDKException('Client', '102', error_message)
                    elif error_code == 1:
                        return True
                elif 'errorMessage' in response.json():
                    error_message = "Failed to get clients for name change operation" \
                                    "with errorCode [{0}], errorMessage [{1}]".format(response.json().get('errorCode'),
                                                                                      response.json().get(
                                                                                          'errorMessage')
                                                                                      )
                    raise SDKException('Client', '102', error_message)
                else:
                    raise SDKException('Response', '102')
            else:
                raise SDKException(
                    'Response', '101', self._update_response_(
                        response.text))
        else:
            raise SDKException(
                'Response',
                '101',
                self._update_response_(
                    response.text))


# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""File for operating on a Salesforce Subclient.

SalesforceSubclient is the only class defined in this file.

SalesforceSubclient:     Derived class from CloudAppsSubclient Base class, representing a
                            salesforce subclient, and to perform operations on that subclient

SalesforceSubclient:

    _get_subclient_properties()               --  Subclient class method overwritten to add
                                                      salesforce subclient properties as well

    _get_subclient_properties_json()          --  gets all the subclient  related properties of
                                                      salesforce subclient.
    enable_files                              --  Enables files option on subclient content

    disable_files                             --  Disables files option on subclient content

    enable_metadata                           --  Enables metadata option on subclient content

    disable_metadata                          --  Disables metadata option on subclient content

    enable_archived_deleted                   --  Enables backup archived and deleted option on subclient content

    disable_archived_deleted                  --  Disables backup archived and deleted  option on subclient content

    browse()                                  --  Browses the salesforce content

    _check_object_in_browse()                 --  internal method to check the object exists
                                                      in browse content

    _restore_salesforce_options_json()        --  internal method for salesforce options json

    _restore_salesforce_destination_json()    --  internal method for salesforce destination option
                                                      json

    restore_to_file_system()                  --  restores the selected content to filesystem

    restore_to_database()                     --  restores the selected content to database

    restore_to_salesforce_from_database()     --  restores the selected content to salesforce from
                                                      database

    restore_to_salesforce_from_media()        --  restores the selected content to salesforce from
                                                      media

    _prepare_salesforce_restore_json()        --  internal method which prepares entire restore
                                                      json for salesforce

"""

from __future__ import unicode_literals

from base64 import b64encode

from ..casubclient import CloudAppsSubclient

from ...client import Client
from ...agent import Agent
from ...instance import Instance
from ...backupsets.cloudapps.salesforce_backupset import SalesforceBackupset
from ...exception import SDKException


class SalesforceSubclient(CloudAppsSubclient):
    """Derived class from CloudAppsSubclient Base class, representing a Salesforce subclient,
        and to perform operations on that subclient."""

    def __init__(self, backupset_object, subclient_name, subclient_id=None):
        """
        Constructor for the class

        Args:
            backupset_object  (object)  -- instance of the Backupset class

            subclient_name    (str)     -- name of the subclient

            subclient_id      (str)     -- id of the subclient

        """
        self.cloud_apps_subclient_prop = {}
        self._files = None
        self._metadata = None
        self._archived_deleted = None
        self._objects = None
        super(SalesforceSubclient, self).__init__(
            backupset_object, subclient_name, subclient_id)

    def _get_subclient_properties(self):
        """Gets the properties of this subclient.

            Raises:
                SDKException:
                    if response is empty

                    if response is not success

        """
        super(SalesforceSubclient, self)._get_subclient_properties()

        if 'cloudAppsSubClientProp' in self._subclient_properties:
            self._cloud_apps_subclient_prop = self._subclient_properties['cloudAppsSubClientProp']
            if 'salesforceSubclient' in self._cloud_apps_subclient_prop:
                sfsubclient = self._cloud_apps_subclient_prop['salesforceSubclient']
                self._objects = sfsubclient.get('backupSfObjects')
                self._files = sfsubclient.get('backupFileObjects')
                self._metadata = sfsubclient.get('backupSFMetadata')
                self._archived_deleted = sfsubclient.get('backupArchivedandDeletedRecs')

    def _get_subclient_properties_json(self):
        """get the all subclient related properties of this subclient.

           Returns:
                dict - all subclient properties put inside a dict

        """
        subclient_json = {
            "subClientProperties": {
                "proxyClient": self._proxyClient,
                "subClientEntity": self._subClientEntity,
                "cloudAppsSubClientProp": self._cloud_apps_subclient_prop,
                "commonProperties": self._commonProperties,
                "contentOperationType": 1
            }
        }

        return subclient_json

    @property
    def objects(self):
        """getter for salesforce files option"""
        return self._objects

    @property
    def files(self):
        """getter for salesforce files option"""
        return self._files

    @property
    def metadata(self):
        """getter for salesforce metadata option """
        return self._metadata

    @property
    def archived_deleted(self):
        """getter for salesfoce backup archived and deleted data"""
        return self._archived_data

    def enable_files(self):
        """
        Enables files option on subclient content
        """
        if not self.files:
            self._set_subclient_properties("_subclient_properties['cloudAppsSubClientProp']\
                                           ['salesforceSubclient']['backupFileObjects']", True)

    def enable_metadata(self):
        """
        Enables metadata option on subclient content
        """
        if not self.metadata:
            self._set_subclient_properties("_subclient_properties['cloudAppsSubClientProp']\
                                           ['salesforceSubclient']['backupSFMetadata']", True)

    def enable_archived_deleted(self):
        """
        Enables backup archived deleted option on subclient content
        """
        if self.archived_deleted:
            self._set_subclient_properties("_subclient_properties['cloudAppsSubClientProp']\
                                           ['salesforceSubclient']['backupArchivedandDeletedRecs']", False)

    def disable_files(self):
        """
        Disables files option on subclient content
        """
        if self.files:
            self._set_subclient_properties("_subclient_properties['cloudAppsSubClientProp']\
                                           ['salesforceSubclient']['backupFileObjects']", False)

    def disable_metadata(self):
        """
        Enables metadata option on subclient content
        """
        if self.metadata:
            self._set_subclient_properties("_subclient_properties['cloudAppsSubClientProp']\
                                           ['salesforceSubclient']['backupSFMetadata']", False)

    def disable_archived_deleted(self):
        """
        Disables backup archived and deleted option on subclient content
        """
        if not self.archived_deleted:
            self._set_subclient_properties("_subclient_properties['cloudAppsSubClientProp']\
                                           ['salesforceSubclient']['backupArchivedandDeletedRecs']", True)

    def check_object_in_browse(self, object_to_restore, browse_data):
        """Check if the particular object is present in browse of the subclient

            Args:
                object_to_restore     (str)   --  folder path whioch has to be restored

                browse_data           (str)   --  list of objects from browse response

            Raises:
                SDKException:
                    if object is not present in browse result
        """
        source_item = None

        if (object_to_restore.find("/Objects") < 0 and
                object_to_restore.find("/") < 0 and
                object_to_restore.find("/Files") < 0):
            restore_object_name = "/Objects/" + object_to_restore
        else:
            restore_object_name = object_to_restore

        for path in browse_data:
            if path.find(restore_object_name) >= 0:
                source_item = path
                break

        if source_item is None:
            raise SDKException('Subclient', '113')

        return restore_object_name

    def _restore_salesforce_options_json(self, value):
        """setter for the salesforce restore  options in restore json
            Raises:
                SDKException:
                    if input value is not dictionary
        """

        if not isinstance(value, dict):
            raise SDKException('Subclient', '101')

        self._salesforce_restore_option_json = {
            "instanceType": 'SALESFORCE',
            "salesforceRestoreOptions": {
                "restoreToFileSystem": value.get("to_fs", True),
                "pathToStoreCsv": value.get("staging_path", '/tmp/'),
                "dependentRestoreLevel": value.get("dependent_level", 0),
                "isMetadataRestore": value.get("is_metadata", False),
                "restoreToSalesforce": value.get("to_cloud", False),
                "restoreFromDatabase": value.get("from_database", False),
                "overrideTable": value.get("override_table", True),
                "syncDatabase": {
                    "dbEnabled": value.get("db_enabled", False),
                    "dbType": value.get("db_type", 'SQLSERVER'),
                    "dbHost": value.get("db_host_name", ''),
                    "dbPort": value.get("db_port", '1433'),
                    "dbName": value.get("db_name", ''),
                    "dbInstance": value.get("db_instance", ''),
                    "dbUserPassword": {
                        "userName": value.get("db_user_name", ''),
                        "password": value.get("db_user_password", '')
                    },

                }

            }
        }

    def _restore_salesforce_destination_json(self, value):
        """setter for  the salesforce destination restore option in restore JSON
            Raises:
                SDKException:
                    if input value is not dictionary
        """

        if not isinstance(value, dict):
            raise SDKException('Subclient', '101')

        self._destination_restore_json = {
            "destClient": {
                "clientName": value.get("dest_client_name", "")
            },
            "destinationInstance": {
                "instanceName": value.get("dest_instance_name", ""),
                "appName": 'Cloud Apps',
                "clientName": value.get("dest_client_name", "")
            },
            "destinationBackupset": {
                "backupsetName": value.get("dest_backupset_name", ""),
                "instanceName": value.get("dest_instance_name", ""),
                "appName": 'Cloud Apps',
                "clientName": value.get("dest_client_name", "")
            },
            "noOfStreams": value.get("streams", 2)
        }

    def restore_to_file_system(
            self,
            objects_to_restore=None,
            destination_client=None,
            sf_options=None):
        """perform restore to file system to the provided path

        Args:
            objects_to_restore  (str)   --  list of objects to restore

            destination_client  (str)   --  destination client name where cloud connector
                                                package exists if this value not provided,
                                                it will automatically use source backup client

            sf_options          (dict)

                destination_path    (str)   :   staging path for salesforce restore data.
                                                    if this value is not provided, uses download
                                                    cache path from source

                dependent_level     (int)   :   restore children based on selected level.
                                                    0   -   no Children
                                                    1   -   immediate children
                                                    -1  -   All children
                    default: 0

                streams             (int)   :   no of streams to use for restore
                    default: 2

                copy_precedence     (int)   :   copy number to use for restore
                    default: 0

                from_time           (str)   :   date to get the contents after
                                                    format: dd/MM/YYYY
                                                    gets content from 01/01/1970 if not specified
                    default: 0

                to_time             (str)   :   date to get the contents before
                                                    format: dd/MM/YYYY
                                                    gets content till latest if not specified
                    default: 0

                show_deleted_files  (bool)  :   include deleted files in the content or not
                    default: True

         Raises:
                SDKException:
                    if from time value is incorrect

                    if to time value is incorrect

                    if to time is less than from time

                    if failed to browse content

                    if response is empty

                    if response is not success

                    if destination client does not exist on commcell

        """

        file_restore_option = {}

        if sf_options is None:
            sf_options = {}

        # check if client name is correct
        if destination_client is None:
            destination_client = self._backupset_object._instance_object.proxy_client

        if isinstance(destination_client, Client):
            client = destination_client
        elif isinstance(destination_client, str):
            client = Client(self._commcell_object, destination_client)
        else:
            raise SDKException('Subclient', '105')

        file_restore_option["client_name"] = client.client_name
        file_restore_option["destination_path"] = sf_options.get(
            "destination_path", self._backupset_object.download_cache_path
        )

        self._restore_destination_json(file_restore_option)

        # process the objects to restore
        if isinstance(objects_to_restore, list):
            objects_to_restore_list = objects_to_restore

        else:
            objects_to_restore_list = [objects_to_restore]

        file_restore_option["paths"] = []
        browse_files, _ = self.browse(
            path='/Objects',
            from_time=sf_options.get("from_time", 0),
            to_time=sf_options.get("to_time", 0)
        )

        for each_object in objects_to_restore_list:
            if each_object.find('/Files') < 0:
                file_restore_option["paths"].append(
                    self.check_object_in_browse("%s" % each_object, browse_files)
                )

        # set the salesforce options
        file_restore_option["staging_path"] = sf_options.get(
            "destination_path",
            self._backupset_object.download_cache_path
        )
        file_restore_option["dependent_level"] = sf_options.get("dependent_level", 0)
        file_restore_option["to_fs"] = True
        file_restore_option["streams"] = sf_options.get("streams", 2)

        # set the browse option
        file_restore_option["copy_precedence_applicable"] = True
        file_restore_option["copy_precedence"] = sf_options.get("copy_precedence", 0)

        # prepare and execute the Json
        request_json = self._prepare_salesforce_restore_json(file_restore_option)

        return self._process_restore_response(request_json)

    def restore_to_database(
            self,
            objects_to_restore=None,
            destination_client=None,
            sf_options=None):
        """perform Restore to  Database

        Args:
            objects_to_restore  (str)   --  list of objects to restore

            destination_client  (str)   --  destination clientname where cloud connector package
                                                exists. if this value not provided, it will
                                                automatically use source backup client

            sf_options          (dict)

                destination_path    (str)   :   staging path for salesforce restore data.
                                                    if this value is not provided, it will
                                                    automatically use download cache path
                                                    from source

                db_type             (str)   :   database type. if database details does not
                                                    provided, it will use syncdb database
                                                    for restore
                    default: SQLSERVER

                db_host_name             (str)   :   database hostname (ex:dbhost.company.com)

                db_instance         (str)   :   database instance name
                                                    (provide if applicable for that database type)

                db_name             (str)   :   database database name
                                                    (it is where data will be imported)

                db_port             (str)   :   database connection port
                    default: 1433

                db_user_name        (str)   :   database username
                                                    (it should have read/write permissions on db)

                db_user_password    (str)   :   database user password

                overrirde_table     (bool)  :   overrides the tables on  database
                    default: True

                dependent_level     (int)   :   restore dependent object based on selected level.
                                                    0   -   no Children
                                                    1   -   immediate children
                                                   -1   -   All children
                    default: 0

                streams             (int)   :   no of streams to use for restore
                    default: 2

                copy_precedence     (int)   :   copy number to use for restore
                    default: 0

                from_date           (str)   :   date to get the contents after
                                                    format: dd/MM/YYYY
                                                    gets contents from 01/01/1970 if not specified
                    default: 0

                to_date             (str)   :   date to get the contents before
                                                    format: dd/MM/YYYY
                                                    gets contents till current day if not specified
                    default: 0

                show_deleted_files  (bool)  :   include deleted files in the content or not
                    default: True

        Raises:
            SDKException:
                if from time value is incorrect

                if to time value is incorrect

                if to time is less than from time

                if failed to browse content

                if response is empty

                if response is not success

                if destination client does not exist on commcell

                if all the database details not provided

        """
        file_restore_option = {}

        if sf_options is None:
            sf_options = {}

        # check if client name is correct
        if destination_client is None:
            destination_client = self._backupset_object._instance_object.proxy_client

        if isinstance(destination_client, Client):
            dest_client = destination_client
        elif isinstance(destination_client, str):
            dest_client = Client(self._commcell_object, destination_client)
        else:
            raise SDKException('Subclient', '105')

        if not ('db_host_name' in sf_options and
                'db_instance' in sf_options and
                'db_name' in sf_options and
                'db_user_name' in sf_options and
                'db_user_password' in sf_options):
            raise SDKException('Salesforce', '101')

        # set the destination client
        file_restore_option["client_name"] = dest_client.client_name
        file_restore_option["destination_path"] = sf_options.get(
            "destination_path", self._backupset_object.download_cache_path
        )

        self._restore_destination_json(file_restore_option)

        # process the objects to restore
        if isinstance(objects_to_restore, list):
            objects_to_restore_list = objects_to_restore

        else:
            objects_to_restore_list = [objects_to_restore]

        file_restore_option["paths"] = []
        browse_files, _ = self.browse(
            path='/Objects',
            from_time=sf_options.get("from_time", 0),
            to_time=sf_options.get("to_time", 0)
        )

        for each_object in objects_to_restore_list:
            if each_object.find('/Files') < 0:
                file_restore_option["paths"].append(
                    self.check_object_in_browse("%s" % each_object, browse_files)
                )

        # set the salesforce options
        file_restore_option["staging_path"] = sf_options.get(
            "destination_path", self._backupset_object.download_cache_path
        )
        file_restore_option["dependent_level"] = sf_options.get("dependent_level", 0)
        file_restore_option["streams"] = sf_options.get("streams", 2)
        file_restore_option["to_fs"] = False
        file_restore_option["db_enabled"] = True
        file_restore_option["db_type"] = sf_options.get("db_type", 'SQLSERVER')
        file_restore_option["db_host_name"] = sf_options.get("db_host_name", "")
        file_restore_option["db_instance"] = sf_options.get("db_instance", "")
        file_restore_option["db_name"] = sf_options.get("db_name", "autorestoredb")
        file_restore_option["db_port"] = sf_options.get("db_port", '1433')
        file_restore_option["db_user_name"] = sf_options.get("db_user_name", 'sa')
        db_base64_password = b64encode(sf_options['db_user_password'].encode()).decode()
        file_restore_option["db_user_password"] = db_base64_password
        file_restore_option["override_table"] = sf_options.get("override_table", True)

        # set the browse option
        file_restore_option["copy_precedence_applicable"] = True
        file_restore_option["copy_precedence"] = sf_options.get("copy_precedence", 0)
        file_restore_option["from_time"] = sf_options.get("from_time", 0)
        file_restore_option["to_time"] = sf_options.get("to_time", 0)

        # prepare and execute the Json
        request_json = self._prepare_salesforce_restore_json(file_restore_option)

        return self._process_restore_response(request_json)

    def restore_to_salesforce_from_database(
            self,
            objects_to_restore=None,
            destination_client=None,
            destination_instance=None,
            destination_backupset=None,
            sf_options=None):
        """perform Restore to Salesforce from Database

        Args:
            objects_to_restore      (str)   --  list of objects to restore

            destination_client      (str)   --  destination pseudo client name.
                                                    if this value not provided, it will
                                                    automatically select source client

            destination_instance    (str)   --  destination instance name.
                                                    if this value not provided, it will
                                                    automatically select source instance name

            destination_backupset   (str)   --  destination backupset name.
                                                    if this value not provided, it will
                                                    automatically select source backupset
            sf_options              (dict)

                destination_path    (str)   :   staging path for salesforce restore data

                db_type             (str)   :   database type. if database details does not
                                                    provided, it will use syncdb database
                                                    for restore
                    default: SQLSERVER

                db_host_name             (str)   :   database hostname (ex:dbhost.company.com)

                db_instance         (str)   :   database instance name
                                                    (provide if applicable for that database type)

                db_name             (str)   :   database database name
                                                    (it is where data will be imported)

                db_port             (str)   :   database connection port
                    default: 1433

                db_user_name        (str)   :   database username
                                                    (read/write permissions needed on db)

                db_user_password    (str)   :   database user password

                overrirde_table     (bool)  :   overrides the tables on  database
                    default: True

                dependent_level     (int)   :   restore children based on selected level.
                                                    0   -   no Children
                                                    1   -   immediate children
                                                    -1  -   All children
                    default: 0

                streams             (int)   :   no of streams to use for restore
                    default: 2

                copy_precedence     (int)   :   copy number to use for restore
                    default: 0

                from_time           (str)   :   date to get the contents after
                                                    format: dd/MM/YYYY
                                                    gets contents from 01/01/1970 if not specified
                    default: None

                to_time             (str)   :   date to get the contents before
                                                    format: dd/MM/YYYY
                                                    gets contents till current day if not specified
                    default: None

                show_deleted_files  (bool)  :   include deleted files in the content or not
                    default: True

        Raises:
            SDKException:
                if from date value is incorrect

                if to date value is incorrect

                if to date is less than from date

                if failed to browse content

                if response is empty

                if response is not success

                if destination client does not exist

                if destination instance does not exist

                if destination backupset does not exist

                if syncdb is not enabled and user not provided the database details

        """
        file_restore_option = {}

        if sf_options is None:
            sf_options = {}

        # check if client name is correct
        if destination_client is None:
            destination_client = self._backupset_object._agent_object._client_object

        if isinstance(destination_client, Client):
            dest_client = destination_client
        elif isinstance(destination_client, str):
            dest_client = Client(self._commcell_object, destination_client)
        else:
            raise SDKException('Subclient', '105')

        dest_agent = Agent(dest_client, 'Cloud Apps', '134')

        # check if instance name is correct
        if destination_instance is None:
            destination_instance = self._backupset_object._instance_object

        if isinstance(destination_instance, Instance):
            dest_instance = destination_instance
        elif isinstance(destination_instance, str):
            dest_instance = dest_agent.instances.get(destination_instance)
        else:
            raise SDKException('Subclient', '113')

        # check if backupset name is correct
        if destination_backupset is None:
            destination_backupset = self._backupset_object

        if isinstance(destination_backupset, SalesforceBackupset):
            dest_backupset = destination_backupset
        elif isinstance(destination_backupset, str):
            dest_backupset = SalesforceBackupset(dest_instance, destination_backupset)
        else:
            raise SDKException('Subclient', '114')

        if not self._backupset_object.is_sync_db_enabled:
            if not (
                    'db_host_name' in sf_options and 'db_instance' in sf_options and
                    'db_name' in sf_options and 'db_user_name' in sf_options and
                    'db_user_password' in sf_options):
                raise SDKException('Salesforce', '101')

        # set salesforce destination client
        file_restore_option["dest_client_name"] = dest_client.client_name
        file_restore_option["dest_instance_name"] = dest_instance.instance_name
        file_restore_option["dest_backupset_name"] = dest_backupset.backupset_name

        self._restore_salesforce_destination_json(file_restore_option)

        # process the objects to restore
        if isinstance(objects_to_restore, list):
            objects_to_restore_list = objects_to_restore

        else:
            objects_to_restore_list = [objects_to_restore]

        file_restore_option["paths"] = []
        browse_files, _ = self.browse(
            path='/Objects', from_time=sf_options.get("from_time", 0),
            to_time=sf_options.get("to_time", 0))

        for each_object in objects_to_restore_list:
            if each_object.find('/Files') < 0:
                file_restore_option["paths"].append(
                    self.check_object_in_browse(
                        "%s" %
                        each_object,
                        browse_files))

        # set the salesforce options
        file_restore_option["staging_path"] = sf_options.get(
            "destination_path",
            dest_backupset.download_cache_path)
        file_restore_option["dependent_level"] = sf_options.get("dependent_level", 0)
        file_restore_option["streams"] = sf_options.get("streams", 2)
        file_restore_option["to_fs"] = False
        file_restore_option["to_cloud"] = True
        file_restore_option["from_database"] = True
        file_restore_option["db_enabled"] = True
        if self._backupset_object.is_sync_db_enabled or ('db_host_name' in sf_options):
            if self._backupset_object.sync_db_type is None:
                dbtype = 'SQLSERVER'
            else:
                dbtype = self._backupset_object.sync_db_type
            file_restore_option["db_type"] = sf_options.get("db_type", dbtype)
            file_restore_option["db_host_name"] = sf_options.get(
                "db_host_name", self._backupset_object.sync_db_host
            )
            file_restore_option["db_instance"] = sf_options.get(
                "db_instance", self._backupset_object.sync_db_instance
            )
            file_restore_option["db_name"] = sf_options.get(
                "db_name", self._backupset_object.sync_db_name
            )
            file_restore_option["db_port"] = sf_options.get(
                "db_port", self._backupset_object.sync_db_port
            )
            file_restore_option["db_user_name"] = sf_options.get(
                "db_user_name", self._backupset_object.sync_db_user_name
            )

            if 'db_user_password' in sf_options:
                sf_options['_db_base64_password'] = b64encode(
                    sf_options['db_user_password'].encode()).decode()

            file_restore_option["db_user_password"] = sf_options.get(
                "_db_base64_password",
                self._backupset_object._sync_db_user_password)
        else:
            raise SDKException('Salesforce', '101')

        file_restore_option["override_table"] = sf_options.get("override_table", True)

        # set the browse option
        file_restore_option["client_name"] = self._backupset_object._agent_object._client_object.client_name
        file_restore_option["copy_precedence_applicable"] = True
        file_restore_option["copy_precedence"] = sf_options.get("copy_precedence", 0)
        file_restore_option["from_time"] = sf_options.get("from_time", 0)
        file_restore_option["to_time"] = sf_options.get("to_time", 0)

        # prepare and execute the Json
        request_json = self._prepare_salesforce_restore_json(file_restore_option)

        return self._process_restore_response(request_json)

    def restore_to_salesforce_from_media(
            self,
            objects_to_restore=None,
            destination_client=None,
            destination_instance=None,
            destination_backupset=None,
            sf_options=None):
        """perform Restore to Salesforce from Media.

        Args:
            objects_to_restore      (str)   --  list of objects to restore

            destination_client      (str)   --  destination pseudo client name.
                                                    if this value not provided, it will
                                                    automatically select source client

            destination_instance    (str)   --  destination instance name.
                                                    if this value not provided, it will
                                                    automatically select source instance name

            destination_backupset   (str)   --  destination backupset name.
                                                    if this value not provided, it will
                                                    automatically select source backupset
            sf_options              (dict)

                destination_path    (str)   :   staging path for salesforce restore data

                db_type             (str)   :   database type. if database details does not
                                                    provided, it will use syncdb database
                                                    for restore
                default: SQLSERVER

                db_host_name             (str)   :   database hostname (ex:dbhost.company.com)

                db_instance         (str)   :   database instance name
                                                    (provide if applicable for that database type)

                db_name             (str)   :   database database name
                                                    (it is where data will be imported)

                db_port             (str)   :   database connection port
                    default: 1433

                db_user_name        (str)   :   database username
                                                    (read/write permissions needed on db)

                db_user_password    (str)   :   database user password

                overrirde_table     (bool)  :   overrides the tables on  database
                    default: True

                dependent_level     (int)   :   restore children based on selected level.
                                                    0   -   no Children
                                                    1   -   immediate children
                                                    -1  -   All children
                    default: 0

                streams             (int)   :   no of streams to use for restore
                    default: 2

                copy_precedence     (int)   :   copy number to use for restore
                    default: 0

                from_time           (str)   :   date to get the contents after
                                                    format: dd/MM/YYYY
                                                    gets contents from 01/01/1970 if not specified
                    default: None

                to_time             (str)   :   date to get the contents before
                                                    format: dd/MM/YYYY
                                                    gets contents till current day if not specified
                    default: None

                show_deleted_files  (bool)  :   include deleted files in the content or not
                    default: True

        Raises:
                SDKException:
                    if from date value is incorrect

                    if to date value is incorrect

                    if to date is less than from date

                    if failed to browse content

                    if response is empty

                    if response is not success

                    if destination client does not exist

                    if destination instance does not exist

                    if destination backupset does not exist

                    if user does not provide staging database details

        """

        file_restore_option = {}

        if sf_options is None:
            sf_options = {}

        # check if client name is correct
        if destination_client is None:
            destination_client = self._backupset_object._agent_object._client_object

        if isinstance(destination_client, Client):
            dest_client = destination_client
        elif isinstance(destination_client, str):
            dest_client = Client(self._commcell_object, destination_client)
        else:
            raise SDKException('Subclient', '105')

        dest_agent = Agent(dest_client, 'Cloud Apps')

        # check if instance name is correct
        if destination_instance is None:
            destination_instance = self._backupset_object._instance_object

        if isinstance(destination_instance, Instance):
            dest_instance = destination_instance
        elif isinstance(destination_instance, str):
            dest_instance = dest_agent.instances.get(destination_instance)
        else:
            raise SDKException('Subclient', '113')

        # check if backupset name is correct
        if destination_backupset is None:
            destination_backupset = self._backupset_object

        if isinstance(destination_backupset, SalesforceBackupset):
            dest_backupset = destination_backupset
        elif isinstance(destination_backupset, str):
            dest_backupset = SalesforceBackupset(dest_instance, destination_backupset)
        else:
            raise SDKException('Subclient', '114')

        if not ('db_host_name' in sf_options and
                'db_instance' in sf_options and
                'db_name' in sf_options and
                'db_user_name' in sf_options and
                'db_user_password' in sf_options):
            raise SDKException('Salesforce', '101')

        file_restore_option["dest_client_name"] = dest_client.client_name
        file_restore_option["dest_instance_name"] = dest_instance.instance_name
        file_restore_option["dest_backupset_name"] = dest_backupset.backupset_name

        self._restore_salesforce_destination_json(file_restore_option)

        # process the objects to restore
        if isinstance(objects_to_restore, list):
            objects_to_restore_list = objects_to_restore

        else:
            objects_to_restore_list = [objects_to_restore]

        file_restore_option["paths"] = []
        browse_files, _ = self.browse(
            path='/Objects',
            from_time=sf_options.get("from_time", 0),
            to_time=sf_options.get("to_time", 0)
        )

        for each_object in objects_to_restore_list:
            if each_object.find('/Files') < 0:
                file_restore_option["paths"].append(
                    self.check_object_in_browse("%s" % each_object, browse_files)
                )

        # set the salesforce options
        file_restore_option["staging_path"] = sf_options.get(
            "destination_path", dest_backupset.download_cache_path
        )
        file_restore_option["dependent_level"] = sf_options.get("dependent_level", 0)
        file_restore_option["streams"] = sf_options.get("streams", 2)
        file_restore_option["to_fs"] = False
        file_restore_option["to_cloud"] = True
        file_restore_option["from_database"] = False
        file_restore_option["db_enabled"] = True
        file_restore_option["db_type"] = sf_options.get("db_type", 'SQLSERVER')
        file_restore_option["db_host_name"] = sf_options.get("db_host_name", "")
        file_restore_option["db_instance"] = sf_options.get("db_instance", "")
        file_restore_option["db_name"] = sf_options.get("db_name", 'autorestoredb')
        file_restore_option["db_port"] = sf_options.get("db_port", '1433')
        file_restore_option["db_user_name"] = sf_options.get("db_user_name", 'sa')
        db_base64_password = b64encode(sf_options['db_user_password'].encode()).decode()
        file_restore_option["db_user_password"] = db_base64_password
        file_restore_option["override_table"] = sf_options.get("override_table", True)

        # set the browse option
        file_restore_option["client_name"] = self._backupset_object._agent_object._client_object.client_name
        file_restore_option["copy_precedence_applicable"] = True
        file_restore_option["copy_precedence"] = sf_options.get("copy_precedence", 0)
        file_restore_option["from_time"] = sf_options.get("from_time", 0)
        file_restore_option["to_time"] = sf_options.get("to_time", 0)

        # prepare and execute the Json
        request_json = self._prepare_salesforce_restore_json(file_restore_option)

        return self._process_restore_response(request_json)

    def metadata_restore_to_salesforce(
            self,
            metadata_list=None,
            destination_client=None,
            destination_instance=None,
            destination_backupset=None,
            **sf_options):
        """perform Restore to Salesforce from Media.

        Args:
            metadata_list           (list)  --  List of metadata components to restore like "folder/component.type"

            destination_client      (str)   --  destination pseudo client name.
                                                    if this value not provided, it will
                                                    automatically select source client

            destination_instance    (str)   --  destination instance name.
                                                    if this value not provided, it will
                                                    automatically select source instance name

            destination_backupset   (str)   --  destination backupset name.
                                                    if this value not provided, it will
                                                    automatically select source backupset
            sf_options:

                destination_path    (str)   :   staging path for salesforce restore data

                streams             (int)   :   no of streams to use for restore
                    default: 2

                copy_precedence     (int)   :   copy number to use for restore
                    default: 0

                from_time           (str)   :   date to get the contents after
                                                    format: dd/MM/YYYY
                                                    gets contents from 01/01/1970 if not specified
                    default: None

                to_time             (str)   :   date to get the contents before
                                                    format: dd/MM/YYYY
                                                    gets contents till current day if not specified
                    default: None

                show_deleted_files  (bool)  :   include deleted files in the content or not
                    default: True

        Raises:
                SDKException:
                    if from date value is incorrect

                    if to date value is incorrect

                    if to date is less than from date

                    if failed to browse content

                    if response is empty

                    if response is not success

                    if destination client does not exist

                    if destination instance does not exist

                    if destination backupset does not exist

        """

        file_restore_option = {}

        if not sf_options:
            sf_options = {}

        # check if client name is correct
        if destination_client is None:
            destination_client = self._backupset_object._agent_object._client_object

        if isinstance(destination_client, Client):
            dest_client = destination_client
        elif isinstance(destination_client, str):
            dest_client = self._commcell_object.clients.get(destination_client)
        else:
            raise SDKException('Subclient', '105')

        dest_agent = Agent(dest_client, 'Cloud Apps')

        # check if instance name is correct
        if destination_instance is None:
            destination_instance = self._backupset_object._instance_object

        if isinstance(destination_instance, Instance):
            dest_instance = destination_instance
        elif isinstance(destination_instance, str):
            dest_instance = dest_agent.instances.get(destination_instance)
        else:
            raise SDKException('Subclient', '113')

        # check if backupset name is correct
        if destination_backupset is None:
            destination_backupset = self._backupset_object

        if isinstance(destination_backupset, SalesforceBackupset):
            dest_backupset = destination_backupset
        elif isinstance(destination_backupset, str):
            dest_backupset = SalesforceBackupset(dest_instance, destination_backupset)
        else:
            raise SDKException('Subclient', '114')

        file_restore_option["dest_client_name"] = dest_client.client_name
        file_restore_option["dest_instance_name"] = dest_instance.instance_name
        file_restore_option["dest_backupset_name"] = dest_backupset.backupset_name

        self._restore_salesforce_destination_json(file_restore_option)

        file_restore_option["paths"] = [f"/Metadata/unpackaged/{metadata_path}" for metadata_path in metadata_list]

        # set the salesforce options
        file_restore_option["staging_path"] = sf_options.get(
            "destination_path", dest_backupset.download_cache_path
        )
        file_restore_option["dependent_level"] = sf_options.get("dependent_level", 0)
        file_restore_option["streams"] = sf_options.get("streams", 2)
        file_restore_option["to_fs"] = False
        file_restore_option["to_cloud"] = True
        file_restore_option["from_database"] = False
        file_restore_option["is_metadata"] = True

        # set the browse option
        file_restore_option["client_name"] = self._backupset_object._agent_object._client_object.client_name
        file_restore_option["copy_precedence_applicable"] = True
        file_restore_option["copy_precedence"] = sf_options.get("copy_precedence", 0)
        file_restore_option["from_time"] = sf_options.get("from_time", 0)
        file_restore_option["to_time"] = sf_options.get("to_time", 0)

        # prepare and execute the Json
        request_json = self._prepare_salesforce_restore_json(file_restore_option)

        return self._process_restore_response(request_json)

    def _prepare_salesforce_restore_json(self, file_restore_option):
        """prepares the  salesforce restore json from getters."""
        self._restore_fileoption_json(file_restore_option)
        self._restore_salesforce_options_json(file_restore_option)
        self._restore_browse_option_json(file_restore_option)
        self._impersonation_json(file_restore_option)
        self._restore_common_options_json(file_restore_option)

        request_json = {
            "taskInfo": {
                "associations": [self._subClientEntity],
                "task": self._json_task,
                "subTasks": [{
                    "subTask": self._json_restore_subtask,
                    "options": {
                        "restoreOptions": {
                            "impersonation": self._impersonation_json_,
                            "cloudAppsRestoreOptions": self._salesforce_restore_option_json,
                            "browseOption": self._browse_restore_json,
                            "commonOptions": self._commonoption_restore_json,
                            "destination": self._destination_restore_json,
                            "fileOption": self._fileoption_restore_json,

                        }
                    }
                }]
            }
        }

        return request_json
